from __future__ import annotations

import logging
import os

import numpy as np
import pathlib
from abc import ABC, abstractmethod

import trimesh
from .tf_transformations import translation_matrix, quaternion_matrix
from trimesh import Trimesh
from typing_extensions import Tuple, Union, Any, List, Optional, Dict, TYPE_CHECKING, Self, Type

import pycrap
import pycrap.ontologies
from pycrap.ontologies import Base, has_child_link, has_parent_link
from .datastructures.dataclasses import JointState, AxisAlignedBoundingBox, Color, LinkState, VisualShape, \
    MeshVisualShape, RotatedBoundingBox
from .datastructures.enums import JointType
from .datastructures.pose import PoseStamped, TransformStamped, Point
from .datastructures.world_entity import WorldEntity, PhysicalBody
from .failures import ObjectDescriptionNotFound, LinkHasNoGeometry, LinkGeometryHasNoMesh
from .local_transformer import LocalTransformer
from .ros import logwarn_once, logwarn

if TYPE_CHECKING:
    from .world_concepts.world_object import Object


class EntityDescription(ABC):
    """
    A description of an entity. This can be a link, joint or object description.
    """

    def __init__(self, parsed_description: Optional[Any] = None):
        """
        :param parsed_description: The parsed description (most likely from a description file) of the entity.
        """
        self.parsed_description = parsed_description

    @property
    @abstractmethod
    def origin(self) -> PoseStamped:
        """
        :return: the origin of this entity.
        """
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """
        :return: the name of this entity.
        """
        pass


class LinkDescription(EntityDescription):
    """
    A link description of an object.
    """

    def __init__(self, parsed_link_description: Any, mesh_dir: Optional[str] = None):
        super().__init__(parsed_link_description)
        self.mesh_dir = mesh_dir

    @property
    @abstractmethod
    def geometry(self) -> Union[List[VisualShape], VisualShape, None]:
        """
        The geometry type of the collision element of this link.
        """
        pass


class JointDescription(EntityDescription):
    """
    A class that represents the description of a joint.
    """

    def __init__(self, parsed_joint_description: Optional[Any] = None, is_virtual: bool = False):
        """
        :param parsed_joint_description: The parsed description of the joint (e.g. from urdf or mjcf file).
        :param is_virtual: True if the joint is virtual (i.e. not a physically existing joint), False otherwise.
        """
        super().__init__(parsed_joint_description)
        self.is_virtual: Optional[bool] = is_virtual

    @property
    @abstractmethod
    def type(self) -> JointType:
        """
        :return: The type of this joint.
        """
        pass

    @property
    @abstractmethod
    def axis(self) -> Point:
        """
        :return: The axis of this joint, for example the rotation axis for a revolute joint.
        """
        pass

    @property
    @abstractmethod
    def has_limits(self) -> bool:
        """
        :return: True if the joint has limits, False otherwise.
        """
        pass

    @property
    def limits(self) -> Tuple[float, float]:
        """
        :return: The lower and upper limits of this joint.
        """
        lower, upper = self.lower_limit, self.upper_limit
        if lower > upper:
            lower, upper = upper, lower
        return lower, upper

    @property
    @abstractmethod
    def lower_limit(self) -> Union[float, None]:
        """
        :return: The lower limit of this joint, or None if the joint has no limits.
        """
        pass

    @property
    @abstractmethod
    def upper_limit(self) -> Union[float, None]:
        """
        :return: The upper limit of this joint, or None if the joint has no limits.
        """
        pass

    @property
    @abstractmethod
    def parent(self) -> str:
        """
        :return: The name of the parent link of this joint.
        """
        pass

    @property
    @abstractmethod
    def child(self) -> str:
        """
        :return: The name of the child link of this joint.
        """
        pass

    @property
    def damping(self) -> float:
        """
        :return: The damping of this joint.
        """
        raise NotImplementedError

    @property
    def friction(self) -> float:
        """
        :return: The friction of this joint.
        """
        raise NotImplementedError


class ObjectEntity:
    """
    An abstract base class that represents a part of an Object.
    This can be a link or a joint of an Object.
    """

    def __init__(self, obj: Object):
        self.object: Object = obj

    @property
    def object_name(self) -> str:
        """
        The name of the object to which this joint belongs.
        """
        return self.object.name

    @property
    def object_id(self) -> int:
        """
        :return: the id of the object to which this entity belongs.
        """
        return self.object.id


class Link(PhysicalBody, ObjectEntity, LinkDescription, ABC):
    """
    A link of an Object in the World.
    """

    def __init__(self, _id: int, link_description: LinkDescription, obj: Object,
                 concept: Type[Base] = pycrap.ontologies.Link, parse_name: bool = True):
        self.description = link_description
        PhysicalBody.__init__(self, _id, obj.world, concept=concept, parse_name=parse_name)
        ObjectEntity.__init__(self, obj)
        LinkDescription.__init__(self, link_description.parsed_description, link_description.mesh_dir)
        self.local_transformer: LocalTransformer = LocalTransformer()
        self.constraint_ids: Dict[Link, int] = {}

    def reset(self):
        """
        Reset the link to its initial state.
        """
        self.constraint_ids = {}
        self.reset_concepts()

    @property
    def parts(self) -> Dict[str, PhysicalBody]:
        return {}

    @property
    def parent_entity(self) -> Object:
        """
        :return: The parent of this link, which is the object.
        """
        return self.object

    @property
    def name(self) -> str:
        """
        :return: The name of this link.
        """
        return self.description.name

    def get_axis_aligned_bounding_box(self, shift_to_link_position: bool = True) -> AxisAlignedBoundingBox:
        """
        :param shift_to_link_position: If True, return the bounding box transformed to the link pose.
        :return: The axis-aligned bounding box of a link. First try to get it from the simulator, if not,
         then calculate it depending on the type of the link geometry.
        """
        try:
            return self.world.get_link_axis_aligned_bounding_box(self)
        except NotImplementedError:
            bounding_box = self.get_axis_aligned_bounding_box_from_geometry()
            if shift_to_link_position:
                return bounding_box.shift_by(self.pose.position)
            else:
                return bounding_box

    def get_rotated_bounding_box(self) -> RotatedBoundingBox:
        """
        :return: The rotated bounding box of a link. First try to get it from the simulator, if not,
         then calculate it depending on the type of the link geometry.
        """
        try:
            return self.world.get_link_rotated_bounding_box(self)
        except NotImplementedError:
            return self.get_axis_aligned_bounding_box_from_geometry().get_rotated_box(self.transform)

    def get_axis_aligned_bounding_box_from_geometry(self) -> AxisAlignedBoundingBox:
        if isinstance(self.geometry, List):
            all_boxes = [geom.get_axis_aligned_bounding_box(self.get_mesh_path([geom])[0])
                         if isinstance(geom, MeshVisualShape) else geom.get_axis_aligned_bounding_box()
                         for geom in self.geometry
                         ]
            bounding_box = AxisAlignedBoundingBox.from_multiple_bounding_boxes(all_boxes)
        else:
            geom = self.geometry
            bounding_box = geom.get_axis_aligned_bounding_box(self.get_mesh_path([geom])[0]) \
                if isinstance(geom, MeshVisualShape) else geom.get_axis_aligned_bounding_box()
        return bounding_box

    def get_convex_hull(self) -> Trimesh:
        """
        :return: The convex hull of the link geometry.
        """
        try:
            return self.world.get_body_convex_hull(self)
        except NotImplementedError:

            if len(self.geometry) > 0 and isinstance(self.geometry[0], MeshVisualShape):
                mesh_paths = self.get_mesh_path(self.geometry)
                meshes = [trimesh.load_mesh(mesh_path) for mesh_path in mesh_paths]
                mesh = meshes[0].union(meshes[1:]) if len(meshes) > 1 else meshes[0]
                translation = translation_matrix(self.transform.translation.to_list())
                rotation = quaternion_matrix(self.transform.rotation.to_list())
                matrix = np.dot(translation, rotation)
                return mesh.convex_hull.apply_transform(matrix)
            else:
                raise LinkGeometryHasNoMesh(self.name, type(self.geometry).__name__)

    def _plot_convex_hull(self):
        """
        Plot the convex hull of the link geometry.
        """
        hull = self.get_convex_hull()
        hull.show()

    def get_mesh_path(self, geometry: Optional[List[MeshVisualShape]] = None) -> List[str]:
        """
        :param geometry: The geometry/geometries for which the mesh path(s) should be returned.
        :return: The path(s) of the mesh file(s) of this link if the geometry is a mesh.
        """
        if geometry is None:
            geometry = self.geometry
        if not hasattr(geometry, "__iter__"):
            geometry = [geometry]
        return [self.get_mesh_filename(geom) for geom in geometry if isinstance(geom, MeshVisualShape)]

    def get_mesh_filename(self, geometry: MeshVisualShape) -> str:
        """
        :return: The mesh file name of this link if the geometry is a mesh, otherwise raise a LinkGeometryHasNoMesh.
        :raises LinkHasNoGeometry: If the link has no geometry.
        :raises LinkGeometryHasNoMesh: If the geometry is not a mesh.
        """
        if geometry is None:
            raise LinkHasNoGeometry(self.name)
        if isinstance(geometry, MeshVisualShape):
            return geometry.file_name
        else:
            raise LinkGeometryHasNoMesh(self.name, type(geometry).__name__)

    def set_object_pose_given_link_pose(self, pose: PoseStamped) -> None:
        """
        Set the pose of this link to the given pose.
        NOTE: This will move the entire object such that the link is at the given pose, it will not consider any joints
        that can allow the link to be at the given pose.

        :param pose: The target pose for this link.
        """
        self.object.set_pose(self.get_object_pose_given_link_pose(pose))

    def get_object_pose_given_link_pose(self, pose):
        """
        Get the object pose given the link pose, which could be a hypothetical link pose to see what would be the object
        pose in that case (assuming that the object itself moved not the joints).

        :param pose: The link pose.
        """
        return (pose.to_transform_stamped(self.tf_frame) * self.get_transform_to_root_link()).to_pose_stamped()

    def get_pose_given_object_pose(self, pose):
        """
        Get the link pose given the object pose, which could be a hypothetical object pose to see what would be the link
        pose in that case (assuming that the object itself moved not the joints).

        :param pose: The object pose.
        """
        return (pose.to_transform_stamped(self.object.tf_frame) * self.get_transform_from_root_link()).to_pose_stamped()

    def get_transform_from_root_link(self) -> TransformStamped:
        """
        Return the transformation from the root link of the object to this link.
        """
        return self.get_transform_from_link(self.object.root_link)

    def get_transform_to_root_link(self) -> TransformStamped:
        """
        Return the transformation from this link to the root link of the object.
        """
        return self.get_transform_to_link(self.object.root_link)

    @property
    def current_state(self) -> LinkState:
        return LinkState(self.body_state, self.constraint_ids.copy())

    @current_state.setter
    def current_state(self, link_state: LinkState) -> None:
        if self.current_state != link_state:
            if not self.all_constraint_links_belong_to_same_world(link_state):
                raise ValueError("All constraint links must belong to the same world, since the constraint ids"
                                 "are unique to the world and cannot be transferred between worlds.")
            self.body_state = link_state.body_state
            self.constraint_ids = link_state.constraint_ids

    def all_constraint_links_belong_to_same_world(self, other: LinkState) -> bool:
        """
        Check if all links belong to the same world as the links in the other link state.

        :param other: The state of the other link.
        :return: True if all links belong to the same world, False otherwise.
        """
        return all([link.world == other_link.world for link, other_link in zip(self.constraint_ids.keys(),
                                                                               other.constraint_ids.keys())])

    def add_fixed_constraint_with_link(self, child_link: Link,
                                       child_to_parent_transform: Optional[TransformStamped] = None) -> int:
        """
        Add a fixed constraint between this link and the given link, to create attachments for example.

        :param child_link: The child link to which a fixed constraint should be added.
        :param child_to_parent_transform: The transformation between the two links.
        :return: The unique id of the constraint.
        """
        if child_to_parent_transform is None:
            child_to_parent_transform = child_link.get_transform_to_link(self)
        constraint_id = self.world.add_fixed_constraint(self, child_link, child_to_parent_transform)
        self.constraint_ids[child_link] = constraint_id
        child_link.constraint_ids[self] = constraint_id
        return constraint_id

    def remove_constraint_with_link(self, child_link: 'Link') -> None:
        """
        Remove the constraint between this link and the given link.

        :param child_link: The child link of the constraint that should be removed.
        """
        self.world.remove_constraint(self.constraint_ids[child_link])
        del self.constraint_ids[child_link]
        if self in child_link.constraint_ids.keys():
            del child_link.constraint_ids[self]

    @property
    def is_only_link(self) -> bool:
        """
        :return: True if this link is the only link, False otherwise.
        """
        return self.object.has_one_link

    @property
    def is_root(self) -> bool:
        """
        :return: True if this link is the root link, False otherwise.
        """
        return self.object.get_root_link_id() == self.id

    def get_transform_to_link(self, link: 'Link') -> TransformStamped:
        """
        :param link: The link to which the transformation should be returned.
        :return: A Transform object with the transformation from this link to the given link.
        """
        return link.get_transform_from_link(self)

    def get_transform_from_link(self, link: 'Link') -> TransformStamped:
        """
        :param link: The link from which the transformation should be returned.
        :return: A Transform object with the transformation from the given link to this link.
        """
        return self.get_pose_wrt_link(link).to_transform_stamped(self.tf_frame)

    def get_pose_wrt_link(self, link: 'Link') -> PoseStamped:
        """
        :param link: The link with respect to which the pose should be returned.
        :return: A Pose object with the pose of this link with respect to the given link.
        """
        return self.local_transformer.transform_pose(self.pose, link.tf_frame)

    def get_origin_transform(self) -> TransformStamped:
        """
        :return: the transformation between the link frame and the origin frame of this link.
        """
        return self.origin.to_transform_stamped(self.tf_frame)

    @property
    def pose(self) -> PoseStamped:
        """
        :return: The pose of this link.
        """
        return self.world.get_link_pose(self)

    @pose.setter
    def pose(self, pose: PoseStamped) -> None:
        logwarn_once("Setting the pose of a link is not allowed,"
                     " change object pose and/or joint position to affect the link pose.")

    @property
    def color(self) -> Color:
        """
        :return: A Color object containing the rgba_color of this link.
        """
        return self.world.get_link_color(self)

    @color.setter
    def color(self, color: Color) -> None:
        """
        Set the color of this link, could be rgb or rgba.

        :param color: The color as a list of floats, either rgb or rgba.
        """
        self.world.set_link_color(self, color)

    @property
    def tf_frame(self) -> str:
        """
        The name of the tf frame of this link.
        """
        return f"{self.object.tf_frame}/{self.name}"

    @property
    def origin_transform(self) -> TransformStamped:
        """
        The transformation between the link frame and the origin frame of this link.
        """
        return self.origin.to_transform_stamped(self.tf_frame)

    def __copy__(self):
        return Link(self.id, self.description, self.object)


class RootLink(Link, ABC):
    """
    The root link of an Object in the World.
    This differs from the normal AbstractLink class in that the pose and the tf_frame is the same as that of the object.
    """

    def __init__(self, obj: Object):
        Link.__init__(self, obj.get_root_link_id(), obj.get_root_link_description(), obj,
                      concept=pycrap.ontologies.RootLink, parse_name=False)

        if not self.world.is_prospection_world:
            self.ontology_individual.is_part_of = [obj.ontology_individual]

    @property
    def tf_frame(self) -> str:
        """
        :return: the tf frame of the root link, which is the same as the tf frame of the object.
        """
        return self.object.tf_frame

    @property
    def pose(self) -> PoseStamped:
        """
        :return: The pose of the root link, which is the same as the pose of the object.
        """
        return self.object.pose

    @pose.setter
    def pose(self, pose: PoseStamped) -> None:
        """
        Set the pose of the root link to the given pose by setting the pose of the object.
        """
        self.object.pose = pose

    def __copy__(self):
        return RootLink(self.object)


class Joint(WorldEntity, ObjectEntity, JointDescription, ABC):
    """
    Represent a joint of an Object in the World.
    """

    def __init__(self, _id: int,
                 joint_description: JointDescription,
                 obj: Object, is_virtual: Optional[bool] = False,
                 concept: Type[Base] = pycrap.ontologies.Joint):
        self.description = joint_description
        WorldEntity.__init__(self, _id, obj.world, concept=concept, parse_name=False)
        ObjectEntity.__init__(self, obj)
        JointDescription.__init__(self, joint_description.parsed_description, is_virtual)

        self.acceptable_error = (self.world.conf.revolute_joint_position_tolerance if self.type == JointType.REVOLUTE
                                 else self.world.conf.prismatic_joint_position_tolerance)
        self._update_position()
        self._update_ontology_data()

    def _update_ontology_data(self):
        """
        Update the ontology data of this joint and its parent and child links.
        """
        if self.world.is_prospection_world:
            return
        self.ontology_individual.is_a = [has_child_link.some(self.child_link.ontology_individual)]
        if self.parent_link.ontology_individual:
            self.ontology_individual.is_a = [has_parent_link.some(self.parent_link.ontology_individual)]
            self.child_link.ontology_individual.is_part_of = [self.parent_link.ontology_individual]

    @property
    def name(self) -> str:
        """
        :return: The name of this joint.
        """
        return self.description.name

    @property
    def parent_entity(self) -> Link:
        """
        :return: The parent of this joint, which is the object.
        """
        return self.parent_link

    @property
    def tf_frame(self) -> str:
        """
        The tf frame of a joint is the tf frame of the child link.
        """
        return self.child_link.tf_frame

    @property
    def pose(self) -> PoseStamped:
        """
        :return: The pose of this joint. The pose is the pose of the child link of this joint.
        """
        return self.child_link.pose

    def _update_position(self) -> None:
        """
        Update the current position of the joint from the physics simulator.
        """
        self._current_position = self.world.get_joint_position(self)

    @property
    def parent_link(self) -> Link:
        """
        :return: The parent link as a AbstractLink object.
        """
        return self.object.get_link(self.parent)

    @property
    def child_link(self) -> Link:
        """
        :return: The child link as a AbstractLink object.
        """
        return self.object.get_link(self.child)

    @property
    def position(self) -> float:
        if self.world.conf.update_poses_from_sim_on_get:
            self._update_position()
        return self._current_position

    def reset_position(self, position: float) -> None:
        self.world.reset_joint_position(self, position)
        self._update_position()

    def get_object_id(self) -> int:
        """
        :return: The integer id of the object to which this joint belongs.
        """
        return self.object.id

    @position.setter
    def position(self, joint_position: float) -> None:
        """
        Set the position of the given joint to the given joint pose. If the pose is outside the joint limits,
         issue a warning. However, set the joint either way.

        :param joint_position: The target pose for this joint
        """
        # TODO Limits for rotational (infinite) joints are 0 and 1, they should be considered separately
        if self.has_limits:
            low_lim, up_lim = self.limits
            if not low_lim <= joint_position <= up_lim:
                logwarn(
                    f"The joint position has to be within the limits of the joint. The joint limits for {self.name}"
                    f" are {low_lim} and {up_lim}")
                logwarn(f"The given joint position was: {joint_position}")
                # Temporarily disabled because kdl outputs values exciting joint limits
                # return
        self.reset_position(joint_position)

    def enable_force_torque_sensor(self) -> None:
        self.world.enable_joint_force_torque_sensor(self.object, self.id)

    def disable_force_torque_sensor(self) -> None:
        self.world.disable_joint_force_torque_sensor(self.object, self.id)

    def get_reaction_force_torque(self) -> List[float]:
        return self.world.get_joint_reaction_force_torque(self.object, self.id)

    def get_applied_motor_torque(self) -> float:
        return self.world.get_applied_joint_motor_torque(self.object, self.id)

    @property
    def current_state(self) -> JointState:
        return JointState(self.position, self.acceptable_error)

    @current_state.setter
    def current_state(self, joint_state: JointState) -> None:
        """
        Update the current state of this joint from the given joint state if the position is different.

        :param joint_state: The joint state to update from.
        """
        if self.current_state != joint_state:
            self.position = joint_state.position

    def __copy__(self):
        return Joint(self.id, self.description, self.object, self.is_virtual)


class ObjectDescription(EntityDescription):
    """
    A class that represents the description of an object.
    """

    mesh_extensions: Tuple[str] = (".obj", ".stl", ".dae", ".ply")
    """
    The file extensions of the mesh files that can be used to generate a description file.
    """

    class Link(Link, ABC):
        ...

    class RootLink(RootLink, ABC):
        ...

    class Joint(Joint, ABC):
        ...

    def __init__(self, path: Optional[str] = None):
        """
        :param path: The path of the file to update the description data from.
        """
        super().__init__(None)
        self._links: Optional[List[LinkDescription]] = None
        self._joints: Optional[List[JointDescription]] = None
        self._link_map: Optional[Dict[str, Any]] = None
        self._joint_map: Optional[Dict[str, Any]] = None
        self.original_path: Optional[str] = path
        if path:
            self.xml_path = path if path.endswith((".xml", ".urdf", ".xml")) else None
            self.update_description_from_file(path)
        else:
            self._parsed_description = None

        self.virtual_joint_names: List[str] = []

    @property
    @abstractmethod
    def child_map(self) -> Dict[str, List[Tuple[str, str]]]:
        """
        :return: A dictionary mapping the name of a link to its children which are represented as a tuple of the child
            joint name and the link name.
        """
        pass

    @property
    @abstractmethod
    def parent_map(self) -> Dict[str, Tuple[str, str]]:
        """
        :return: A dictionary mapping the name of a link to its parent joint and link as a tuple.
        """
        pass

    @property
    @abstractmethod
    def link_map(self) -> Dict[str, LinkDescription]:
        """
        :return: A dictionary mapping the name of a link to its description.
        """
        pass

    @property
    @abstractmethod
    def joint_map(self) -> Dict[str, JointDescription]:
        """
        :return: A dictionary mapping the name of a joint to its description.
        """
        pass

    def is_joint_virtual(self, name: str) -> bool:
        """
        :param name: The name of the joint.
        :return: True if the joint is virtual, False otherwise.
        """
        return name in self.virtual_joint_names

    @abstractmethod
    def add_joint(self, name: str, child: str, joint_type: JointType,
                  axis: Point, parent: Optional[str] = None, origin: Optional[PoseStamped] = None,
                  lower_limit: Optional[float] = None, upper_limit: Optional[float] = None,
                  is_virtual: Optional[bool] = False) -> None:
        """
        Add a joint to this object.

        :param name: The name of the joint.
        :param child: The name of the child link.
        :param joint_type: The type of the joint.
        :param axis: The axis of the joint.
        :param parent: The name of the parent link.
        :param origin: The origin of the joint.
        :param lower_limit: The lower limit of the joint.
        :param upper_limit: The upper limit of the joint.
        :param is_virtual: True if the joint is virtual, False otherwise.
        """
        pass

    @abstractmethod
    def merge_description(self, other: ObjectDescription, parent_link: Optional[str] = None,
                          child_link: Optional[str] = None,
                          joint_type: JointType = JointType.FIXED,
                          axis: Optional[Point] = None,
                          lower_limit: Optional[float] = None, upper_limit: Optional[float] = None,
                          child_pose_wrt_parent: Optional[PoseStamped] = None,
                          in_place: bool = False,
                          new_description_file: Optional[str] = None) -> Union[ObjectDescription, Self]:
        """
        Merge the description of this object with the description of the other object.

        :param other: The object description to merge with this one.
        :param parent_link: The name of the parent link of the joint connecting the two objects.
        :param child_link: The name of the child link of the joint connecting the two objects.
        :param joint_type: The type of the joint connecting the two objects.
        :param axis: The axis of the joint connecting the two objects.
        :param lower_limit: The lower limit of the joint connecting the two objects.
        :param upper_limit: The upper limit of the joint connecting the two objects.
        :param child_pose_wrt_parent: The pose of the child link with respect to the parent link.
        :param in_place: True if the merge should be done in place, False otherwise.
        :param new_description_file: If given, the new description will be saved to this file, otherwise the new
            description will be saved in place of the original file.
        :return: The merged object description, could be a new object description if in_place is False else self.
        """
        pass

    def update_description_from_file(self, path: str) -> None:
        """
        Update the description of this object from the file at the given path.

        :param path: The path of the file to update from.
        """
        self.xml_path = path
        self._parsed_description = self.load_description(path)

    def update_description_from_string(self, description_string: str) -> None:
        """
        Update the description of this object from the given description string.

        :param description_string: The description string to update from.
        """
        self._parsed_description = self.load_description_from_string(description_string)

    def load_description_from_string(self, description_string: str) -> Any:
        """
        Load the description from the given string.

        :param description_string: The description string to load from.
        """
        raise NotImplementedError

    @property
    def parsed_description(self) -> Any:
        """
        :return: The object parsed from the description file.
        """
        return self._parsed_description

    @parsed_description.setter
    def parsed_description(self, parsed_description: Any):
        """
        :param parsed_description: The parsed description object (depends on the description file type).
        """
        self._parsed_description = parsed_description

    @abstractmethod
    def load_description(self, path: str) -> Any:
        """
        Load the description from the file at the given path.

        :param path: The path to the source file, if only a filename is provided then the resources directories will be
         searched.
        """
        pass

    def generate_description_from_file(self, path: str, name: str, extension: str, save_path: str,
                                       scale_mesh: Optional[float] = None,
                                       mesh_transform: Optional[TransformStamped] = None,
                                       color: Optional[Color] = None) -> None:
        """
        Generate and preprocess the description from the file at the given path and save the preprocessed
        description. The generated description will be saved at the given save path.

        :param path: The path of the file to preprocess.
        :param name: The name of the object.
        :param extension: The file extension of the file to preprocess.
        :param save_path: The path to save the generated description file.
        :param scale_mesh: The scale of the mesh.
        :param mesh_transform: The transformation matrix to apply to the mesh.
        :param color: The color of the object.
        :raises ObjectDescriptionNotFound: If the description file could not be found/read.
        """

        if extension in self.mesh_extensions:
            if extension == ".ply":
                mesh = trimesh.load(path)
                if scale_mesh is not None:
                    mesh.apply_scale(scale_mesh)
                if mesh_transform is not None:
                    transform = mesh_transform.get_homogeneous_matrix()
                    mesh.apply_transform(transform)

                root_dir = os.path.dirname(path)
                object_name = os.path.basename(path).split('.')[0]

                new_dir_name = f"converted_{object_name}"
                path = os.path.join(root_dir, new_dir_name, object_name+".obj")

                if not os.path.exists(path):
                    os.makedirs(os.path.dirname(path), exist_ok=True)
                mesh.export(path)
            self.generate_from_mesh_file(path, name, save_path=save_path, color=color)
        elif extension == self.get_file_extension():
            self.generate_from_description_file(path, save_path=save_path)
        else:
            try:
                # Using the description from the parameter server
                self.generate_from_parameter_server(path, save_path=save_path)
            except KeyError:
                logging.warning(f"Couldn't find file data in the ROS parameter server")

        if not self.check_description_file_exists_and_can_be_read(save_path):
            raise ObjectDescriptionNotFound(name, path, extension)

    @staticmethod
    def check_description_file_exists_and_can_be_read(path: str) -> bool:
        """
        Check if the description file exists at the given path.

        :param path: The path to the description file.
        :return: True if the file exists, False otherwise.
        """
        exists = os.path.exists(path)
        if exists:
            with open(path, "r") as file:
                exists = bool(file.read())
        return exists

    @staticmethod
    def write_description_to_file(description_string: str, save_path: str) -> None:
        """
        Write the description string to the file at the given path.

        :param description_string: The description string to write.
        :param save_path: The path of the file to write to.
        """
        with open(save_path, "w") as file:
            file.write(description_string)

    def get_file_name(self, path_object: pathlib.Path, extension: str, object_name: str) -> str:
        """
        :param path_object: The path object of the description file or the mesh file.
        :param extension: The file extension of the description file or the mesh file.
        :param object_name: The name of the object.
        :return: The file name of the description file.
        """
        if extension in self.mesh_extensions:
            file_name = path_object.stem + self.get_file_extension()
        elif extension == self.get_file_extension():
            file_name = path_object.name
        else:
            file_name = object_name + self.get_file_extension()

        return file_name

    @classmethod
    @abstractmethod
    def generate_from_mesh_file(cls, path: str, name: str, save_path: str, color: Color) -> None:
        """
        Generate a description file from one of the mesh types defined in the mesh_extensions and
        return the path of the generated file. The generated file will be saved at the given save_path.

        :param path: The path to the .obj file.
        :param name: The name of the object.
        :param save_path: The path to save the generated description file.
        :param color: The color of the object.
        """
        pass

    @classmethod
    @abstractmethod
    def generate_from_description_file(cls, path: str, save_path: str, make_mesh_paths_absolute: bool = True) -> None:
        """
        Preprocess the given file and return the preprocessed description string. The preprocessed description will be
        saved at the given save_path.

        :param path: The path of the file to preprocess.
        :param save_path: The path to save the preprocessed description file.
        :param make_mesh_paths_absolute: Whether to make the mesh paths absolute.
        """
        pass

    @classmethod
    @abstractmethod
    def generate_from_parameter_server(cls, name: str, save_path: str) -> None:
        """
        Preprocess the description from the ROS parameter server and return the preprocessed description string.
        The preprocessed description will be saved at the given save_path.

        :param name: The name of the description on the parameter server.
        :param save_path: The path to save the preprocessed description file.
        """
        pass

    @property
    @abstractmethod
    def links(self) -> List[LinkDescription]:
        """
        :return: A list of links descriptions of this object.
        """
        pass

    def get_link_by_name(self, link_name: str) -> LinkDescription:
        """
        :return: The link description with the given name.
        """
        return self.link_map[link_name]

    @property
    @abstractmethod
    def joints(self) -> List[JointDescription]:
        """
        :return: A list of joints descriptions of this object.
        """
        pass

    def get_joint_by_name(self, joint_name: str) -> JointDescription:
        """
        :return: The joint description with the given name.
        """
        return self.joint_map[joint_name]

    @abstractmethod
    def get_root(self) -> str:
        """
        :return: the name of the root link of this object.
        """
        pass

    def get_tip(self) -> str:
        """
        :return: the name of the tip link of this object.
        """
        raise NotImplementedError

    @abstractmethod
    def get_chain(self, start_link_name: str, end_link_name: str, joints: Optional[bool] = True,
                  links: Optional[bool] = True, fixed: Optional[bool] = True) -> List[str]:
        """
        :return: the chain of links from 'start_link_name' to 'end_link_name'.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_file_extension() -> str:
        """
        :return: The file extension of the description file.
        """
        pass



from threading import Lock
import numpy as np

from .default_process_modules import *
from .. import world_reasoning as btr
from ..utils import _apply_ik
from ..designators.motion_designator import *
from ..datastructures.enums import JointType, StaticJointState
from ..external_interfaces.ik import request_ik

from ..datastructures.world import World
from ..local_transformer import LocalTransformer
from ..process_module import ProcessModule, ProcessModuleManager
from ..robot_description import RobotDescription
from ..world_concepts.world_object import Object

def _park_arms(arm):
    """
    Defines the joint poses for the parking positions of the arm of Donbot and applies them to the
    in the World defined robot.
    :return: None
    """

    robot = World.robot
    if arm == "left":
        for joint, pose in RobotDescription.current_robot_description.get_static_joint_chain("left",
                                                                                             StaticJointState.Park).items():
            robot.set_joint_position(joint, pose)

class BoxyParkArms(ProcessModule):
    """
    This process module is for moving the arms in a parking position.
    It is currently not used.
    """

    def _execute(self, desig):
        solutions = desig.reference()
        if solutions['cmd'] == 'park':
            _park_arms()


class BoxyMoveHead(ProcessModule):
    """
    This process module moves the head to look at a specific point in the world coordinate frame.
    This point can either be a position or an object.
    """

    def _execute(self, desig):
        target = desig.target
        robot = World.robot

        local_transformer = LocalTransformer()

        pose_in_shoulder = local_transformer.transform_pose(target, robot.get_link_tf_frame("neck_shoulder_link"))

        if pose_in_shoulder.position.x >= 0 and pose_in_shoulder.position.x >= abs(pose_in_shoulder.position.y):
            robot.set_multiple_joint_positions(
                RobotDescription.current_robot_description.get_static_joint_chain("neck", "front"))
        if pose_in_shoulder.position.y >= 0 and pose_in_shoulder.position.y >= abs(pose_in_shoulder.position.x):
            robot.set_multiple_joint_positions(
                RobotDescription.current_robot_description.get_static_joint_chain("neck", "neck_right"))
        if pose_in_shoulder.position.x <= 0 and abs(pose_in_shoulder.position.x) > abs(pose_in_shoulder.position.y):
            robot.set_multiple_joint_positions(
                RobotDescription.current_robot_description.get_static_joint_chain("neck", "back"))
        if pose_in_shoulder.position.y <= 0 and abs(pose_in_shoulder.position.y) > abs(pose_in_shoulder.position.x):
            robot.set_multiple_joint_positions(
                RobotDescription.current_robot_description.get_static_joint_chain("neck", "neck_left"))

        pose_in_shoulder = local_transformer.transform_pose(target, robot.get_link_tf_frame("neck_shoulder_link"))

        new_pan = np.arctan2(pose_in_shoulder.position.y, pose_in_shoulder.position.x)

        robot.set_joint_position("neck_shoulder_pan_joint",
                                 new_pan + robot.get_joint_position("neck_shoulder_pan_joint"))


class BoxyDetecting(ProcessModule):
    """
    This process module tries to detect an object with the given type. To be detected the object has to be in
    the field of view of the robot.
    """

    def _execute(self, desig):
        robot = World.robot
        object_type = desig.object_type
        # Should be "wide_stereo_optical_frame"
        cam_link_name = RobotDescription.current_robot_description.get_camera_link()
        # should be [0, 0, 1]
        front_facing_axis = RobotDescription.current_robot_description.get_default_camera().front_facing_axis

        objects = World.current_world.get_object_by_type(object_type)
        for obj in objects:
            if btr.visible(obj, robot.get_link_pose(cam_link_name), front_facing_axis):
                return obj


class BoxyManager(DefaultManager):

    def __init__(self):
        super().__init__()
        self.robot_name = "boxy"

    def looking(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return BoxyMoveHead(self._looking_lock)

    def detecting(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return BoxyDetecting(self._detecting_lock)

BoxyManager()



from ..robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, \
    RobotDescriptionManager, CameraDescription
from ..datastructures.enums import Arms, Grasp, GripperState, GripperType, TorsoState, StaticJointState
from ..ros import get_ros_package_path

filename = get_ros_package_path('pycram') + '/resources/robots/' + "iCub3" + '.urdf'

icub_description = RobotDescription("iCub3", "base_footprint", "torso_1", "torso_pitch",
                                    filename)

################################## Left Arm ##################################
left_arm = KinematicChainDescription("left", "root_link", "l_hand",
                                     icub_description.urdf_object, arm_type=Arms.LEFT)

left_arm.add_static_joint_states(StaticJointState.Park, {"torso_roll": 0,
                                          "torso_pitch": 0,
                                          "torso_yaw": 0,
                                          "l_shoulder_pitch": 0,
                                          "l_shoulder_roll": 0,
                                          "l_shoulder_yaw": 0,
                                          "l_elbow": 0,
                                          "l_wrist_prosup": 0,
                                          "l_wrist_pitch": 0,
                                          "l_wrist_yaw": 0})

icub_description.add_kinematic_chain_description(left_arm)

################################## Left Gripper ##################################
left_gripper = EndEffectorDescription("left_gripper", "l_hand", "l_gripper_tool_frame",
                                      icub_description.urdf_object)
left_gripper.add_static_joint_states(GripperState.OPEN, {"l_hand_thumb_0_joint": 0.0,
                                                         "l_hand_thumb_1_joint": 0.0,
                                                         "l_hand_thumb_2_joint": 0.0,
                                                         "l_hand_thumb_3_joint": 0.0,
                                                         "l_hand_index_0_joint": 0.0,
                                                         "l_hand_index_1_joint": 0.0,
                                                         "l_hand_index_2_joint": 0.0,
                                                         "l_hand_index_3_joint": 0.0,
                                                         "l_hand_middle_0_joint": 0.0,
                                                         "l_hand_middle_1_joint": 0.0,
                                                         "l_hand_middle_2_joint": 0.0,
                                                         "l_hand_middle_3_joint": 0.0,
                                                         "l_hand_ring_0_joint": 0.0,
                                                         "l_hand_ring_1_joint": 0.0,
                                                         "l_hand_ring_2_joint": 0.0,
                                                         "l_hand_ring_3_joint": 0.0,
                                                         "l_hand_little_0_joint": 0.0,
                                                         "l_hand_little_1_joint": 0.0,
                                                         "l_hand_little_2_joint": 0.0,
                                                         "l_hand_little_3_joint": 0.0})
left_gripper.add_static_joint_states(GripperState.CLOSE, {"l_hand_thumb_0_joint": 1.5707963267948966,
                                                          "l_hand_thumb_1_joint": 1.5707963267948966,
                                                          "l_hand_thumb_2_joint": 1.5707963267948966,
                                                          "l_hand_thumb_3_joint": 1.5707963267948966,
                                                          "l_hand_index_0_joint": -0.3490658503988659,
                                                          "l_hand_index_1_joint": 1.5707963267948966,
                                                          "l_hand_index_2_joint": 1.5707963267948966,
                                                          "l_hand_index_3_joint": 1.5707963267948966,
                                                          "l_hand_middle_0_joint": 0.3490658503988659,
                                                          "l_hand_middle_1_joint": 1.5707963267948966,
                                                          "l_hand_middle_2_joint": 1.5707963267948966,
                                                          "l_hand_middle_3_joint": 1.5707963267948966,
                                                          "l_hand_ring_0_joint": 0.3490658503988659,
                                                          "l_hand_ring_1_joint": 1.5707963267948966,
                                                          "l_hand_ring_2_joint": 1.5707963267948966,
                                                          "l_hand_ring_3_joint": 1.5707963267948966,
                                                          "l_hand_little_0_joint": 0.3490658503988659,
                                                          "l_hand_little_1_joint": 1.5707963267948966,
                                                          "l_hand_little_2_joint": 1.5707963267948966,
                                                          "l_hand_little_3_joint": 1.5707963267948966})

left_gripper.end_effector_type = GripperType.FINGER
# left_gripper.opening_distance = 0.548
left_arm.end_effector = left_gripper

################################## Right Arm ##################################
right_arm = KinematicChainDescription("right", "root_link", "r_hand",
                                      icub_description.urdf_object, arm_type=Arms.RIGHT)

right_arm.add_static_joint_states(StaticJointState.Park, {"torso_roll": 0,
                                           "torso_pitch": 0,
                                           "torso_yaw": 0,
                                           "r_shoulder_pitch": 0,
                                           "r_shoulder_roll": 0,
                                           "r_shoulder_yaw": 0,
                                           "r_elbow": 0,
                                           "r_wrist_prosup": 0,
                                           "r_wrist_pitch": 0,
                                           "r_wrist_yaw": 0})

icub_description.add_kinematic_chain_description(right_arm)

################################## Right Gripper ##################################
right_gripper = EndEffectorDescription("right_gripper", "r_hand", "r_gripper_tool_frame",
                                       icub_description.urdf_object)
right_gripper.add_static_joint_states(GripperState.OPEN, {"r_hand_thumb_0_joint": 0.0,
                                                          "r_hand_thumb_1_joint": 0.0,
                                                          "r_hand_thumb_2_joint": 0.0,
                                                          "r_hand_thumb_3_joint": 0.0,
                                                          "r_hand_index_0_joint": 0.0,
                                                          "r_hand_index_1_joint": 0.0,
                                                          "r_hand_index_2_joint": 0.0,
                                                          "r_hand_index_3_joint": 0.0,
                                                          "r_hand_middle_0_joint": 0.0,
                                                          "r_hand_middle_1_joint": 0.0,
                                                          "r_hand_middle_2_joint": 0.0,
                                                          "r_hand_middle_3_joint": 0.0,
                                                          "r_hand_ring_0_joint": 0.0,
                                                          "r_hand_ring_1_joint": 0.0,
                                                          "r_hand_ring_2_joint": 0.0,
                                                          "r_hand_ring_3_joint": 0.0,
                                                          "r_hand_little_0_joint": 0.0,
                                                          "r_hand_little_1_joint": 0.0,
                                                          "r_hand_little_2_joint": 0.0,
                                                          "r_hand_little_3_joint": 0.0})
right_gripper.add_static_joint_states(GripperState.CLOSE, {"r_hand_thumb_0_joint": 1.5707963267948966,
                                                          "r_hand_thumb_1_joint": 1.5707963267948966,
                                                          "r_hand_thumb_2_joint": 1.5707963267948966,
                                                          "r_hand_thumb_3_joint": 1.5707963267948966,
                                                          "r_hand_index_0_joint": -0.3490658503988659,
                                                          "r_hand_index_1_joint": 1.5707963267948966,
                                                          "r_hand_index_2_joint": 1.5707963267948966,
                                                          "r_hand_index_3_joint": 1.5707963267948966,
                                                          "r_hand_middle_0_joint": 0.3490658503988659,
                                                          "r_hand_middle_1_joint": 1.5707963267948966,
                                                          "r_hand_middle_2_joint": 1.5707963267948966,
                                                          "r_hand_middle_3_joint": 1.5707963267948966,
                                                          "r_hand_ring_0_joint": 0.3490658503988659,
                                                          "r_hand_ring_1_joint": 1.5707963267948966,
                                                          "r_hand_ring_2_joint": 1.5707963267948966,
                                                          "r_hand_ring_3_joint": 1.5707963267948966,
                                                          "r_hand_little_0_joint": 0.3490658503988659,
                                                          "r_hand_little_1_joint": 1.5707963267948966,
                                                          "r_hand_little_2_joint": 1.5707963267948966,
                                                          "r_hand_little_3_joint": 1.5707963267948966})

right_gripper.end_effector_type = GripperType.FINGER
# right_gripper.opening_distance = 0.548
right_arm.end_effector = right_gripper

################################## Torso ##################################
# redo to use knees instead of torso
torso = KinematicChainDescription("torso", "root_link", "chest",
                                  icub_description.urdf_object)

torso.add_static_joint_states(TorsoState.HIGH, {"torso_roll": 0})

torso.add_static_joint_states(TorsoState.MID, {"torso_roll": 0})

torso.add_static_joint_states(TorsoState.LOW, {"torso_roll": 0})

icub_description.add_kinematic_chain_description(torso)

################################## Camera ##################################
# real camera unknown at the moment of writing (also missing in urdf), so using dummy camera for now
camera = CameraDescription("eye_camera", "head", 1.27,
                           1.85, 0.99483, 0.75049,
                           [1, 0, 0])
icub_description.add_camera_description(camera)

################################## Neck ##################################
icub_description.add_kinematic_chain("neck", "chest", "head")
icub_description.set_neck(yaw_joint="neck_yaw", pitch_joint="neck_pitch", roll_joint="neck_roll")

################################# Grasps ##################################
left_orientation = [0.5, 0.5, 0.5, 0.5]
left_gripper.update_all_grasp_orientations(left_orientation)

right_orientation = [0, 0, -0.707, 0.707]
right_gripper.update_all_grasp_orientations(right_orientation)

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(icub_description)



from __future__ import annotations

import itertools
import math
from abc import ABC, abstractmethod
from copy import deepcopy, copy
from dataclasses import dataclass, fields, field
from enum import Enum

import numpy as np
import plotly.graph_objects as go
import trimesh
from matplotlib import pyplot as plt

from random_events.interval import closed, SimpleInterval, Bound
from random_events.product_algebra import SimpleEvent, Event
from random_events.variable import Continuous
from typing_extensions import List, Optional, Tuple, Callable, Dict, Any, Union, TYPE_CHECKING, Sequence, Self, \
    deprecated, Type

from pycrap.ontologies import PhysicalObject
from .enums import JointType, Shape, VirtualMobileBaseJointName, Grasp, AxisIdentifier
from .pose import PoseStamped, Point, TransformStamped
from ..ros import logwarn, logwarn_once
from ..utils import classproperty
from ..validation.error_checkers import calculate_joint_position_error, is_error_acceptable

if TYPE_CHECKING:
    from ..description import Link, ObjectDescription
    from ..world_concepts.world_object import Object
    from ..world_concepts.constraints import Attachment
    from .world_entity import PhysicalBody
    from .world import World


@dataclass
class ManipulatorData:
    """
    A dataclass for storing the information of a manipulator that is used for creating a robot description for that
    manipulator. A manipulator is an Arm with an end-effector that can be used to interact with the environment.
    """
    name: str
    """
    Name of the Manipulator.
    """
    base_link: str
    """
    Manipulator's base link.
    """
    arm_end_link: str
    """
    Manipulator's arm end link.
    """
    joint_names: List[str]
    """
    List of joint names.
    """
    home_joint_values: List[float]
    """
    List of joint values for the home position. (default position)
    """
    gripper_name: str
    """
    Name of the gripper at the end of the arm.
    """
    gripper_tool_frame: str
    """
    Name of the frame of the gripper tool.
    """
    gripper_joint_names: List[str]
    """
    List of gripper joint names.
    """
    closed_joint_values: List[float]
    """
    List of joint values for the gripper in the closed position.
    """
    open_joint_values: List[float]
    """
    List of joint values for the gripper in the open position.
    """
    opening_distance: float
    """
    The opening distance of the gripper.
    """
    fingers_link_names: Optional[List[str]] = None
    """
    List of link names for the fingers of the gripper.
    """
    relative_dir: str = ''
    """
    Relative directory of the manipulator description file in the resources directory.
    """
    gripper_cmd_topic: Optional[str] = None
    """
    Gripper command topic in ROS if it has one.
    """
    gripper_open_cmd_value: Optional[float] = None
    """
    Grip open command value.
    """
    gripper_close_cmd_value: Optional[float] = None
    """
    Grip close command value.
    """
    gripper_relative_dir: Optional[str] = None
    """
    Relative directory of the gripper description file in the resources directory if it has one and is not part of the
     manipulator description file.
    """


def get_point_as_list(point: Point) -> List[float]:
    """
    Return the point as a list.

    :param point: The point.
    :return: The point as a list
    """
    return [point.x, point.y, point.z]


@dataclass
class Color:
    """
    Dataclass for storing rgba_color as an RGBA value.
    The values are stored as floats between 0 and 1.
    The default rgba_color is white. 'A' stands for the opacity.
    """
    R: float = 1
    G: float = 1
    B: float = 1
    A: float = 1

    @classmethod
    def from_list(cls, color: List[float]):
        """
        Set the rgba_color from a list of RGBA values.

        :param color: The list of RGBA values
        """
        if len(color) == 3:
            return cls.from_rgb(color)
        elif len(color) == 4:
            return cls.from_rgba(color)
        else:
            raise ValueError("Color list must have 3 or 4 elements")

    @classmethod
    def from_rgb(cls, rgb: List[float]):
        """
        Set the rgba_color from a list of RGB values.

        :param rgb: The list of RGB values
        """
        return cls(rgb[0], rgb[1], rgb[2], 1)

    @classmethod
    def from_rgba(cls, rgba: List[float]):
        """
        Set the rgba_color from a list of RGBA values.

        :param rgba: The list of RGBA values
        """
        return cls(rgba[0], rgba[1], rgba[2], rgba[3])

    def get_rgba(self) -> List[float]:
        """
        Return the rgba_color as a list of RGBA values.

        :return: The rgba_color as a list of RGBA values
        """
        return [self.R, self.G, self.B, self.A]

    def get_rgb(self) -> List[float]:
        """
        Return the rgba_color as a list of RGB values.

        :return: The rgba_color as a list of RGB values
        """
        return [self.R, self.G, self.B]


class Colors(Color, Enum):
    """
    Enum for easy access to some common colors.
    """
    PINK = (1, 0, 1, 1)
    BLACK = (0, 0, 0, 1)
    WHITE = (1, 1, 1, 1)
    RED = (1, 0, 0, 1)
    GREEN = (0, 1, 0, 1)
    BLUE = (0, 0, 1, 1)
    YELLOW = (1, 1, 0, 1)
    CYAN = (0, 1, 1, 1)
    MAGENTA = (1, 0, 1, 1)
    GREY = (0.5, 0.5, 0.5, 1)

    @classmethod
    def from_string(cls, color: str) -> Color:
        """
        Set the rgba_color from a string. If the string is not a valid color, it will return the color WHITE.

        :param color: The string of the color
        """
        try:
            return cls[color.upper()]
        except KeyError:
            return cls.WHITE


@dataclass
class BoundingBox:
    """
    Dataclass for storing an axis-aligned bounding box.

    An axis aligned bounding box is the cartesian product of the three closed intervals
    [min_x, max_x] x [min_y, max_y] x [min_z, max_z].

    Depth is the distance between the min_x and max_x, and should always be the long side (excluding height).
    Width is the distance between the min_y and max_y, and should always be the short side (excluding height).
    Height is the distance between the min_z and max_z, "up" is according to how the object would stand on a table.

    Set-Algebraic operations are possible by converting the bounding box to a random event.
    """

    min_x: float
    """
    The minimum x-coordinate of the bounding box.
    """

    min_y: float
    """
    The minimum y-coordinate of the bounding box.
    """

    min_z: float
    """
    The minimum z-coordinate of the bounding box.
    """

    max_x: float
    """
    The maximum x-coordinate of the bounding box.
    """

    max_y: float
    """
    The maximum y-coordinate of the bounding box.
    """

    max_z: float
    """
    The maximum z-coordinate of the bounding box.
    """

    x_variable = Continuous("x")
    """
    The x variable for the random-events interface.
    """

    y_variable = Continuous("y")
    """
    The y variable for the random-events interface.
    """

    z_variable = Continuous("z")
    """
    The z variable for the random-events interface.
    """

    def __hash__(self):
        # The hash should be this since comparing those via hash is checking if those are the same and not just equal
        return id(self)

    @property
    def x_interval(self) -> SimpleInterval:
        """
        :return: The x interval of the bounding box.
        """
        return SimpleInterval(self.min_x, self.max_x, Bound.CLOSED, Bound.CLOSED)

    @property
    def y_interval(self) -> SimpleInterval:
        """
        :return: The y interval of the bounding box.
        """
        return SimpleInterval(self.min_y, self.max_y, Bound.CLOSED, Bound.CLOSED)

    @property
    def z_interval(self) -> SimpleInterval:
        """
        :return: The z interval of the bounding box.
        """
        return SimpleInterval(self.min_z, self.max_z, Bound.CLOSED, Bound.CLOSED)

    @property
    def simple_event(self) -> SimpleEvent:
        """
        :return: The bounding box as a random event.
        """
        return SimpleEvent({self.x_variable: self.x_interval,
                            self.y_variable: self.y_interval,
                            self.z_variable: self.z_interval})

    @classmethod
    def from_simple_event(cls, simple_event: SimpleEvent):
        """
        Create a list of bounding boxes from a simple random event.

        :param simple_event: The random event.
        :return: The list of bounding boxes.
        """
        result = []
        for x, y, z in itertools.product(simple_event[cls.x_variable].simple_sets,
                                         simple_event[cls.y_variable].simple_sets,
                                         simple_event[cls.z_variable].simple_sets):
            result.append(cls(x.lower, y.lower, z.lower, x.upper, y.upper, z.upper))
        return result

    @classmethod
    def from_event(cls, event: Event) -> List[Self]:
        """
        Create a list of bounding boxes from a random event.

        :param event: The random event.
        :return: The list of bounding boxes.
        """
        return [box for simple_event in event.simple_sets for box in cls.from_simple_event(simple_event)]

    def intersection_with(self, other: BoundingBox) -> Optional[BoundingBox]:
        """
        Compute the intersection of two bounding boxes.

        :param other: The other bounding box.
        :return: The intersection of the two bounding boxes or None if they do not intersect.
        """
        result = self.simple_event.intersection_with(other.simple_event)
        if result.is_empty():
            return None
        return self.__class__.from_simple_event(result)[0]

    def contains(self, x: float, y: float, z: float):
        """
        Check if the bounding box contains a point.

        :param x: The x-coordinate of the point.
        :param y: The y-coordinate of the point.
        :param z: The z-coordinate of the point.
        :return: True if the bounding box contains the point, False otherwise.
        """
        return self.simple_event.contains((x, y, z))

    def contains_box(self, other: BoundingBox):
        """
        Check if the bounding box contains another bounding box.

        :param other: The other bounding box.
        :return: True if the bounding box contains the other bounding box, False otherwise.
        """
        return (other.simple_event.as_composite_set() - self.simple_event.as_composite_set()).is_empty()

    @classmethod
    def merge_multiple_bounding_boxes_into_mesh(cls, bounding_boxes: List[BoundingBox],
                                                save_mesh_to: Optional[str] = None,
                                                use_random_events: bool = True,
                                                plot: bool = False) -> trimesh.Trimesh:
        """
        Merge multiple axis-aligned bounding boxes into a single mesh.

        :param bounding_boxes: The list of axis-aligned bounding boxes.
        :param save_mesh_to: The file path to save the mesh to.
        :param use_random_events: If True, use random events to compute the new shape, otherwise use pyvista.
        :param plot: If True, plot the mesh.
        :return: The mesh of the merged bounding boxes.
        """
        if use_random_events:

            all_intervals = [(box.get_min(), box.get_max()) for box in bounding_boxes]
            event = None
            for min_point, max_point in all_intervals:
                new_event = SimpleEvent({cls.x_variable: closed(min_point[0], max_point[0]),
                                         cls.y_variable: closed(min_point[1], max_point[1]),
                                         cls.z_variable: closed(min_point[2], max_point[2])}).as_composite_set()
                # TODO fix this when random events is fixed.
                if event:
                    event = event.__deepcopy__().union_with(new_event)
                else:
                    event = new_event
            if plot:
                fig = go.Figure(event.plot(), event.plotly_layout())
                fig.update_layout(title="Merged Bounding Boxes")
                fig.show()
            mesh = BoundingBox.get_mesh_from_event(event)
        else:
            mesh = BoundingBox.get_mesh_from_boxes(bounding_boxes)
        if plot:
            mesh.show()
            BoundingBox.plot_3d_points([mesh.vertices])
        if save_mesh_to is not None:
            mesh.export(save_mesh_to)
        return mesh

    @property
    @abstractmethod
    def transform(self) -> TransformStamped:
        """
        Get the transformation of the bounding box.
        """
        pass

    def extents(self) -> np.ndarray:
        """
        :return: The size of the bounding box in each dimension.
        """
        return np.array([self.depth, self.width, self.height])

    @staticmethod
    def get_mesh_from_boxes(boxes: List[BoundingBox]) -> trimesh.Trimesh:
        """
        Get the mesh from the boxes

        :param boxes: The list of boxes
        :return: The mesh.
        """
        first_box_mesh = boxes[0].as_mesh
        if len(boxes) == 1:
            return first_box_mesh
        else:
            return first_box_mesh.union([box.as_mesh for box in boxes[1:]]).convex_hull

    @property
    def as_mesh(self) -> trimesh.Trimesh:
        """
        :return: The mesh of the bounding box.
        """
        return trimesh.primitives.Box(self.extents(), self.transform.transform.to_matrix())

    @staticmethod
    def get_mesh_from_event(event: Event) -> trimesh.Trimesh:
        """
        Get the mesh from the event.

        :param event: The event.
        :return: The mesh.
        """
        # form cartesian product of all intervals
        intervals = [value.simple_sets for simple_event in event.simple_sets for _, value in simple_event.items()]
        simple_events = list(itertools.product(*intervals))

        # for every atomic interval
        all_vertices = []
        all_faces = []
        for i, simple_event in enumerate(simple_events):
            x, y, z = 0, 1, 2
            for j in range(2):
                x, y, z = x + j * 3, y + j * 3, z + j * 3
                # Create a 3D mesh trace for the rectangle
                all_vertices.extend([[simple_event[x].lower, simple_event[y].lower, simple_event[z].lower],
                                     [simple_event[x].lower, simple_event[y].lower, simple_event[z].upper],
                                     [simple_event[x].lower, simple_event[y].upper, simple_event[z].lower],
                                     [simple_event[x].lower, simple_event[y].upper, simple_event[z].upper],
                                     [simple_event[x].upper, simple_event[y].lower, simple_event[z].lower],
                                     [simple_event[x].upper, simple_event[y].lower, simple_event[z].upper],
                                     [simple_event[x].upper, simple_event[y].upper, simple_event[z].lower],
                                     [simple_event[x].upper, simple_event[y].upper, simple_event[z].upper]])
                all_faces.extend((np.array(BoundingBox.get_box_faces()) + i * 16 + j * 8).tolist())
        return trimesh.Trimesh(np.array(all_vertices), np.array(all_faces))

    @staticmethod
    def get_box_faces() -> List[List[int]]:
        return [[0, 1, 2], [2, 3, 1], [4, 5, 6], [6, 7, 5],
                [0, 1, 4], [4, 5, 1], [2, 3, 6], [6, 7, 2],
                [0, 2, 4], [4, 6, 2], [1, 3, 5], [5, 7, 3]]

    @classmethod
    def from_min_max(cls, min_point: Sequence[float], max_point: Sequence[float]):
        """
        Set the axis-aligned bounding box from a minimum and maximum point.

        :param min_point: The minimum point
        :param max_point: The maximum point
        """
        return cls(min_point[0], min_point[1], min_point[2], max_point[0], max_point[1], max_point[2])

    @property
    def origin(self) -> List[float]:
        return [(self.min_x + self.max_x) / 2, (self.min_y + self.max_y) / 2, (self.min_z + self.max_z) / 2]

    @property
    def base_origin(self) -> List[float]:
        center = self.origin
        return [center[0], center[1], self.min_z]

    @property
    def origin_point(self) -> Point:
        return Point(**dict(zip(["x", "y", "z"], self.origin)))

    def get_points_list(self) -> List[List[float]]:
        """
        :return: The points of the bounding box as a list of lists of floats.
        """
        return [[point.x, point.y, point.z] for point in self.get_points()]

    def get_points(self) -> List[Point]:
        """
        :return: The points of the bounding box as a list of Point instances.
        """
        return [Point(x=self.min_x, y=self.min_y, z=self.min_z),
                Point(x=self.min_x, y=self.min_y, z=self.max_z),
                Point(x=self.min_x, y=self.max_y, z=self.min_z),
                Point(x=self.min_x, y=self.max_y, z=self.max_z),
                Point(x=self.max_x, y=self.min_y, z=self.min_z),
                Point(x=self.max_x, y=self.min_y, z=self.max_z),
                Point(x=self.max_x, y=self.max_y, z=self.min_z),
                Point(x=self.max_x, y=self.max_y, z=self.max_z)]

    def get_min_max_points(self) -> Tuple[Point, Point]:
        """
        :return: The axis-aligned bounding box as a tuple of minimum and maximum points
        """
        return self.get_min_point(), self.get_max_point()

    def get_min_point(self) -> Point:
        """
        :return: The axis-aligned bounding box as a minimum point
        """
        return Point(x=self.min_x, y=self.min_y, z=self.min_z)

    def get_max_point(self) -> Point:
        """
        :return: The axis-aligned bounding box as a maximum point
        """
        return Point(x=self.max_x, y=self.max_y, z=self.max_z)

    def get_min_max(self) -> Tuple[List[float], List[float]]:
        """
        :return: The axis-aligned bounding box as a tuple of minimum and maximum points
        """
        return self.get_min(), self.get_max()

    def get_min(self) -> List[float]:
        """
        :return: The minimum point of the axis-aligned bounding box
        """
        return [self.min_x, self.min_y, self.min_z]

    def get_max(self) -> List[float]:
        """
        :return: The maximum point of the axis-aligned bounding box
        """
        return [self.max_x, self.max_y, self.max_z]

    def enlarge(self, min_x: float = 0., min_y: float = 0, min_z: float = 0,
                max_x: float = 0., max_y: float = 0., max_z: float = 0.):
        """
        Enlarge the axis-aligned bounding box by a given amount in-place.
        :param min_x: The amount to enlarge the minimum x-coordinate
        :param min_y: The amount to enlarge the minimum y-coordinate
        :param min_z: The amount to enlarge the minimum z-coordinate
        :param max_x: The amount to enlarge the maximum x-coordinate
        :param max_y: The amount to enlarge the maximum y-coordinate
        :param max_z: The amount to enlarge the maximum z-coordinate
        """
        self.min_x -= min_x
        self.min_y -= min_y
        self.min_z -= min_z
        self.max_x += max_x
        self.max_y += max_y
        self.max_z += max_z

    def enlarge_all(self, amount: float):
        """
        Enlarge the axis-aligned bounding box in all dimensions by a given amount in-place.

        :param amount: The amount to enlarge the bounding box
        """
        self.enlarge(amount, amount, amount,
                     amount, amount, amount)

    @property
    def depth(self) -> float:
        return self.max_x - self.min_x

    @property
    def height(self) -> float:
        return self.max_z - self.min_z

    @property
    def width(self) -> float:
        return self.max_y - self.min_y

    @property
    def dimensions(self) -> List[float]:
        """
        According to the IAI conventions, found at https://ai.uni-bremen.de/wiki/3dmodeling/items
        1. z is height, according to how the object would stand on a table
        2. x is depth, representing the long side of the object
        3. y is width, representing the remaining dimension
        """
        if self.width > self.depth:
            logwarn_once("The width of the bounding box is greater than the depth. This means the object's"
                         "axis alignment is potentially going against IAI conventions.")
        return [self.depth, self.width, self.height]

    @staticmethod
    def plot_3d_points(list_of_points: List[np.ndarray]):
        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')

        for points in list_of_points:
            color = np.random.rand(3, )
            ax.scatter(points[:, 0], points[:, 1], points[:, 2], c=color, marker='o')

        ax.set_xlabel('X Label')
        ax.set_ylabel('Y Label')
        ax.set_zlabel('Z Label')
        plt.xlim(0, 2)
        plt.ylim(0, 2)
        ax.set_zlim(0, 2)

        plt.show()


@dataclass
class AxisAlignedBoundingBox(BoundingBox):

    @property
    def transform(self) -> TransformStamped:
        return TransformStamped.from_list(self.origin)

    def get_rotated_box(self, transform: TransformStamped) -> RotatedBoundingBox:
        """
        Apply a transformation to the axis-aligned bounding box and return the transformed axis-aligned bounding box.

        :return: The transformed axis-aligned bounding box
        """
        return RotatedBoundingBox.from_min_max(self.get_min(), self.get_max(), transform)

    @classmethod
    def from_origin_and_half_extents(cls, origin: Point, half_extents: Point):
        """
        Set the axis-aligned bounding box from the origin of the body and half the size.

        :param origin: The origin point
        :param half_extents: The half size of the bounding box.
        """
        min_point = [origin.x - half_extents.x, origin.y - half_extents.y, origin.z - half_extents.z]
        max_point = [origin.x + half_extents.x, origin.y + half_extents.y, origin.z + half_extents.z]
        return cls.from_min_max(min_point, max_point)

    @classmethod
    def from_multiple_bounding_boxes(cls, bounding_boxes: List[AxisAlignedBoundingBox]) -> AxisAlignedBoundingBox:
        """
        Set the axis-aligned bounding box from multiple axis-aligned bounding boxes.

        :param bounding_boxes: The list of axis-aligned bounding boxes.
        """
        min_x = min([box.min_x for box in bounding_boxes])
        min_y = min([box.min_y for box in bounding_boxes])
        min_z = min([box.min_z for box in bounding_boxes])
        max_x = max([box.max_x for box in bounding_boxes])
        max_y = max([box.max_y for box in bounding_boxes])
        max_z = max([box.max_z for box in bounding_boxes])
        return cls(min_x, min_y, min_z, max_x, max_y, max_z)

    def shift_by(self, shift: Point) -> AxisAlignedBoundingBox:
        """
        Shift the axis-aligned bounding box by a given shift.

        :param shift: The shift to apply
        :return: The shifted axis-aligned bounding box
        """
        return AxisAlignedBoundingBox(self.min_x + shift.x, self.min_y + shift.y, self.min_z + shift.z,
                                      self.max_x + shift.x, self.max_y + shift.y, self.max_z + shift.z)


@dataclass
class RotatedBoundingBox(BoundingBox):
    """
    Dataclass for storing a rotated bounding box.
    """

    def __init__(self, min_x: float, min_y: float, min_z: float, max_x: float, max_y: float, max_z: float,
                 transform: Optional[TransformStamped] = None, points: Optional[List[Point]] = None):
        """
        Set the rotated bounding box from a minimum and maximum point.
        :param transform: The transformation
        :param points: The points of the rotated bounding box.
        """
        self._transform: Optional[TransformStamped] = transform
        super().__init__(min_x, min_y, min_z, max_x, max_y, max_z)
        self._points: Optional[List[Point]] = points

    @property
    def transform(self) -> TransformStamped:
        return self._transform

    @classmethod
    def from_min_max(cls, min_point: Sequence[float], max_point: Sequence[float],
                     transform: Optional[TransformStamped] = None):
        """
        Set the rotated bounding box from a minimum, maximum point, and a transformation.

        :param min_point: The minimum point
        :param max_point: The maximum point
        :param transform: The transformation
        """
        return cls(min_point[0], min_point[1], min_point[2], max_point[0], max_point[1], max_point[2], transform)

    def get_points(self) -> List[Point]:
        """
        :return: The points of the rotated bounding box.
        """
        points_array = np.array([[point.x, point.y, point.z] for point in super().get_points()])
        if self._points is None:
            transformed_points = self.transform.apply_transform_to_array_of_points(points_array).tolist()
            self._points = [Point(**dict(zip(["x", "y", "z"], point))) for point in transformed_points]
        return self._points


@dataclass
class CollisionCallbacks:
    """
    Dataclass for storing the collision callbacks which are callables that get called when there is a collision
    or when a collision is no longer there.
    """
    on_collision_cb: Callable
    no_collision_cb: Optional[Callable] = None


@dataclass
class MultiBody:
    """
    Dataclass for storing the information of a multibody which consists of a base and multiple links with joints.
    """
    base_visual_shape_index: int
    base_pose: PoseStamped
    link_visual_shape_indices: List[int]
    link_poses: List[PoseStamped]
    link_masses: List[float]
    link_inertial_frame_poses: List[PoseStamped]
    link_parent_indices: List[int]
    link_joint_types: List[JointType]
    link_joint_axis: List[Point]
    link_collision_shape_indices: List[int]


@dataclass
class VisualShape(ABC):
    """
    Abstract dataclass for storing the information of a visual shape.
    """
    rgba_color: Color
    visual_frame_position: List[float]

    @abstractmethod
    def shape_data(self) -> Dict[str, Any]:
        """
        :return: the shape data of the visual shape (e.g. half extents for a box, radius for a sphere) as a dictionary.
        """
        pass

    @property
    @abstractmethod
    def visual_geometry_type(self) -> Shape:
        """
        :return: The visual geometry type of the visual shape (e.g. box, sphere) as a Shape object.
        """
        pass

    def get_axis_aligned_bounding_box(self) -> AxisAlignedBoundingBox:
        """
        :return: The axis-aligned bounding box of the visual shape.
        """
        raise NotImplementedError


@dataclass
class BoxVisualShape(VisualShape):
    """
    Dataclass for storing the information of a box visual shape
    """
    half_extents: List[float]

    def shape_data(self) -> Dict[str, List[float]]:
        return {"halfExtents": self.half_extents}

    @property
    def visual_geometry_type(self) -> Shape:
        return Shape.BOX

    @property
    def size(self) -> List[float]:
        return self.half_extents

    def get_axis_aligned_bounding_box(self) -> AxisAlignedBoundingBox:
        """
        :return: The axis-aligned bounding box of the box visual shape.
        """
        return AxisAlignedBoundingBox(-self.half_extents[0], -self.half_extents[1], -self.half_extents[2],
                                      self.half_extents[0], self.half_extents[1], self.half_extents[2])


@dataclass
class SphereVisualShape(VisualShape):
    """
    Dataclass for storing the information of a sphere visual shape
    """
    radius: float

    def shape_data(self) -> Dict[str, float]:
        return {"radius": self.radius}

    @property
    def visual_geometry_type(self) -> Shape:
        return Shape.SPHERE

    def get_axis_aligned_bounding_box(self) -> AxisAlignedBoundingBox:
        """
        :return: The axis-aligned bounding box of the sphere visual shape.
        """
        return AxisAlignedBoundingBox(-self.radius, -self.radius, -self.radius, self.radius, self.radius, self.radius)


@dataclass
class CapsuleVisualShape(VisualShape):
    """
    Dataclass for storing the information of a capsule visual shape
    """
    radius: float
    length: float

    def shape_data(self) -> Dict[str, float]:
        return {"radius": self.radius, "length": self.length}

    @property
    def visual_geometry_type(self) -> Shape:
        return Shape.CAPSULE

    def get_axis_aligned_bounding_box(self) -> AxisAlignedBoundingBox:
        """
        :return: The axis-aligned bounding box of the capsule visual shape.
        """
        return AxisAlignedBoundingBox(-self.radius, -self.radius, -self.length / 2,
                                      self.radius, self.radius, self.length / 2)


@dataclass
class CylinderVisualShape(CapsuleVisualShape):
    """
    Dataclass for storing the information of a cylinder visual shape
    """

    @property
    def visual_geometry_type(self) -> Shape:
        return Shape.CYLINDER


@dataclass
class MeshVisualShape(VisualShape):
    """
    Dataclass for storing the information of a mesh visual shape
    """
    scale: List[float]
    file_name: str

    def shape_data(self) -> Dict[str, Union[List[float], str]]:
        return {"meshScale": self.scale, "meshFileName": self.file_name}

    @property
    def visual_geometry_type(self) -> Shape:
        return Shape.MESH

    def get_axis_aligned_bounding_box(self, file_path: Optional[str] = None) -> AxisAlignedBoundingBox:
        """
        :param file_path: An alternative file path.
        :return: The axis-aligned bounding box of the mesh visual shape.
        """
        mesh_file_path = file_path if file_path is not None else self.file_name
        mesh = trimesh.load(mesh_file_path)
        min_bound, max_bound = mesh.bounds
        return AxisAlignedBoundingBox.from_min_max(min_bound, max_bound)


@dataclass
class PlaneVisualShape(VisualShape):
    """
    Dataclass for storing the information of a plane visual shape
    """
    normal: List[float]

    def shape_data(self) -> Dict[str, List[float]]:
        return {"normal": self.normal}

    @property
    def visual_geometry_type(self) -> Shape:
        return Shape.PLANE


VisualShapeUnion = Union[BoxVisualShape, SphereVisualShape, CapsuleVisualShape,
CylinderVisualShape, MeshVisualShape, PlaneVisualShape]


@dataclass
class State(ABC):
    """
    Abstract dataclass for storing the state of an entity (e.g. world, object, link, joint).
    """
    pass


@dataclass
class PhysicalBodyState(State):
    """
    Dataclass for storing the state of a physical body.
    """
    pose: PoseStamped
    is_translating: bool
    is_rotating: bool
    velocity: List[float]
    acceptable_pose_error: Tuple[float, float] = (0.001, 0.001)
    acceptable_velocity_error: Tuple[float, float] = (0.001, 0.001)

    def __eq__(self, other: PhysicalBodyState):
        return (self.pose_is_almost_equal(other)
                and self.is_translating == other.is_translating
                and self.is_rotating == other.is_rotating
                and self.velocity_is_almost_equal(other)
                )

    def pose_is_almost_equal(self, other: PhysicalBodyState) -> bool:
        """
        Check if the pose of the object is almost equal to the pose of another object.

        :param other: The state of the other object.
        :return: True if the poses are almost equal, False otherwise.
        """
        return self.pose.almost_equal(other.pose, other.acceptable_pose_error[0], other.acceptable_pose_error[1])

    def velocity_is_almost_equal(self, other: PhysicalBodyState) -> bool:
        """
        Check if the velocity of the object is almost equal to the velocity of another object.

        :param other: The state of the other object.
        :return: True if the velocities are almost equal, False otherwise.
        """
        if self.velocity is None or other.velocity is None:
            return self.velocity == other.velocity
        return (self.vector_is_almost_equal(self.velocity[:3], other.velocity[:3], self.acceptable_velocity_error[0])
                and self.vector_is_almost_equal(self.velocity[3:], other.velocity[3:],
                                                self.acceptable_velocity_error[1]))

    @staticmethod
    def vector_is_almost_equal(vector1: List[float], vector2: List[float], acceptable_error: float) -> bool:
        """
        Check if the vector is almost equal to another vector.

        :param vector1: The first vector.
        :param vector2: The second vector.
        :param acceptable_error: The acceptable error.
        :return: True if the vectors are almost equal, False otherwise.
        """
        return np.all(np.array(vector1) - np.array(vector2) <= acceptable_error)

    def __copy__(self):
        return PhysicalBodyState(pose=self.pose.copy(),
                                 is_translating=self.is_translating, is_rotating=self.is_rotating,
                                 velocity=self.velocity.copy(),
                                 acceptable_pose_error=deepcopy(self.acceptable_pose_error),
                                 acceptable_velocity_error=deepcopy(self.acceptable_velocity_error))


@dataclass
class LinkState(State):
    """
    Dataclass for storing the state of a link.
    """
    body_state: PhysicalBodyState
    constraint_ids: Dict[Link, int]

    def __eq__(self, other: LinkState):
        return (self.body_state == other.body_state
                and self.all_constraints_exist(other)
                and self.all_constraints_are_equal(other))

    def all_constraints_exist(self, other: LinkState) -> bool:
        """
        Check if all constraints exist in the other link state.

        :param other: The state of the other link.
        :return: True if all constraints exist, False otherwise.
        """
        return (all([cid_k in other.constraint_ids.keys() for cid_k in self.constraint_ids.keys()])
                and len(self.constraint_ids.keys()) == len(other.constraint_ids.keys()))

    def all_constraints_are_equal(self, other: LinkState) -> bool:
        """
        Check if all constraints are equal to the ones in the other link state.

        :param other: The state of the other link.
        :return: True if all constraints are equal, False otherwise.
        """
        return all([cid == other_cid for cid, other_cid in zip(self.constraint_ids.values(),
                                                               other.constraint_ids.values())])

    def __copy__(self):
        return LinkState(copy(self.body_state), constraint_ids=copy(self.constraint_ids))


@dataclass
class JointState(State):
    """
    Dataclass for storing the state of a joint.
    """
    position: float
    acceptable_error: float

    def __eq__(self, other: JointState):
        error = calculate_joint_position_error(self.position, other.position)
        return is_error_acceptable(error, other.acceptable_error)

    def __copy__(self):
        return JointState(position=self.position, acceptable_error=self.acceptable_error)


@dataclass
class ObjectState(State):
    """
    Dataclass for storing the state of an object.
    """
    body_state: PhysicalBodyState
    attachments: Dict[Object, Attachment]
    link_states: Dict[int, LinkState]
    joint_states: Dict[int, JointState]

    def __eq__(self, other: ObjectState):
        return (self.body_state == other.body_state
                and self.all_attachments_exist(other) and self.all_attachments_are_equal(other)
                and self.link_states == other.link_states
                and self.joint_states == other.joint_states)

    @property
    def pose(self) -> PoseStamped:
        return self.body_state.pose

    def all_attachments_exist(self, other: ObjectState) -> bool:
        """
        Check if all attachments exist in the other object state.

        :param other: The state of the other object.
        :return: True if all attachments exist, False otherwise.
        """
        return (all([att_k in other.attachments.keys() for att_k in self.attachments.keys()])
                and len(self.attachments.keys()) == len(other.attachments.keys()))

    def all_attachments_are_equal(self, other: ObjectState) -> bool:
        """
        Check if all attachments are equal to the ones in the other object state.

        :param other: The state of the other object.
        :return: True if all attachments are equal, False otherwise
        """
        return all([att == other_att for att, other_att in zip(self.attachments.values(), other.attachments.values())])

    def __copy__(self):
        return ObjectState(body_state=self.body_state.copy(), attachments=copy(self.attachments),
                           link_states=copy(self.link_states),
                           joint_states=copy(self.joint_states))


@dataclass
class WorldState(State):
    """
    Dataclass for storing the state of the world.
    """
    object_states: Dict[str, ObjectState]
    simulator_state_id: Optional[int] = None

    def __eq__(self, other: WorldState):
        return (self.simulator_state_is_equal(other) and self.all_objects_exist(other)
                and self.all_objects_states_are_equal(other))

    def simulator_state_is_equal(self, other: WorldState) -> bool:
        """
        Check if the simulator state is equal to the simulator state of another world state.

        :param other: The state of the other world.
        :return: True if the simulator states are equal, False otherwise.
        """
        return self.simulator_state_id == other.simulator_state_id

    def all_objects_exist(self, other: WorldState) -> bool:
        """
        Check if all objects exist in the other world state.

        :param other: The state of the other world.
        :return: True if all objects exist, False otherwise.
        """
        return (all([obj_name in other.object_states.keys() for obj_name in self.object_states.keys()])
                and len(self.object_states.keys()) == len(other.object_states.keys()))

    def all_objects_states_are_equal(self, other: WorldState) -> bool:
        """
        Check if all object states are equal to the ones in the other world state.

        :param other: The state of the other world.
        :return: True if all object states are equal, False otherwise.
        """
        return all([obj_state == other_obj_state
                    for obj_state, other_obj_state in zip(self.object_states.values(),
                                                          other.object_states.values())])

    def __copy__(self):
        return WorldState(object_states=deepcopy(self.object_states),
                          simulator_state_id=self.simulator_state_id
                          )


@dataclass
class LateralFriction:
    """
    Dataclass for storing the information of the lateral friction.
    """
    lateral_friction: float
    lateral_friction_direction: List[float]


@dataclass
class ContactPoint:
    """
    Dataclass for storing the information of a contact point between two bodies.
    """
    body_a: PhysicalBody
    body_b: PhysicalBody
    position_on_body_a: Optional[List[float]] = None
    position_on_body_b: Optional[List[float]] = None
    normal_on_body_b: Optional[List[float]] = None  # the contact normal vector on object b pointing towards object a
    distance: Optional[float] = None  # distance between the two objects (+ve for separation, -ve for penetration)
    normal_force: Optional[float] = None  # normal force applied during last step simulation
    lateral_friction_1: Optional[LateralFriction] = None
    lateral_friction_2: Optional[LateralFriction] = None

    @property
    def normal(self) -> List[float]:
        return self.normal_on_body_b

    @property
    def bodies(self) -> Tuple[PhysicalBody, PhysicalBody]:
        return self.body_a, self.body_b

    def __str__(self):
        return f"ContactPoint: {self.body_a.name} - {self.body_b.name}"

    def __repr__(self):
        return self.__str__()


ClosestPoint = ContactPoint
"""
The closest point between two objects which has the same structure as ContactPoint.
"""


class ContactPointsList(list):
    """
    A list of contact points.
    """

    def get_bodies_that_got_removed(self, previous_points: ContactPointsList) -> List[PhysicalBody]:
        """
        Return the bodies that are not in the current points list but were in the initial points list.

        :param previous_points: The initial points list.
        :return: A list of bodies that got removed.
        """
        initial_bodies_in_contact = previous_points.get_all_bodies()
        current_bodies_in_contact = self.get_all_bodies()
        return [body for body in initial_bodies_in_contact if body not in current_bodies_in_contact]

    def get_all_bodies(self, excluded: List[PhysicalBody] = None) -> List[PhysicalBody]:
        """
        :return: A list of all involved bodies in the points.
        """
        excluded = excluded if excluded is not None else []
        return list(set([body for point in self for body in point.bodies if body not in excluded]))

    def check_if_two_objects_are_in_contact(self, obj_a: Object, obj_b: Object) -> bool:
        """
        Check if two objects are in contact.

        :param obj_a: An instance of the Object class that represents the first object.
        :param obj_b: An instance of the Object class that represents the second object.
        :return: True if the objects are in contact, False otherwise.
        """
        return (any([self.is_body_in_object(point.body_b, obj_b)
                     and self.is_body_in_object(point.body_a, obj_a) for point in self]) or
                any([self.is_body_in_object(point.body_a, obj_b)
                     and self.is_body_in_object(point.body_b, obj_a) for point in self]))

    @staticmethod
    def is_body_in_object(body: PhysicalBody, obj: Object) -> bool:
        """
        Check if the body belongs to the object.

        :param body: The body.
        :param obj: The object.
        :return: True if the body belongs to the object, False otherwise.
        """
        return body in list(obj.links.values()) or body == obj

    def get_normals_of_object(self, obj: Object) -> List[List[float]]:
        """
        Get the normals of the object.

        :param obj: An instance of the Object class that represents the object.
        :return: A list of float vectors that represent the normals of the object.
        """
        return self.get_points_of_object(obj).get_normals()

    def get_normals(self) -> List[List[float]]:
        """
        Get the normals of the points.

        :return: A list of float vectors that represent the normals of the contact points.
        """
        return [point.normal_on_body_b for point in self]

    def get_links_in_contact_of_object(self, obj: Object) -> List[PhysicalBody]:
        """
        Get the links in contact of the object.

        :param obj: An instance of the Object class that represents the object.
        :return: A list of Link instances that represent the links in contact of the object.
        """
        return [point.body_b for point in self if point.body_b.parent_entity == obj]

    def get_points_of_object(self, obj: Object) -> ContactPointsList:
        """
        Get the points of the object.

        :param obj: An instance of the Object class that represents the object that the points are related to.
        :return: A ContactPointsList instance that represents the contact points of the object.
        """
        return ContactPointsList([point for point in self if self.is_body_in_object(point.body_b, obj)])

    def get_points_of_link(self, link: Link) -> ContactPointsList:
        """
        Get the points of the link.

        :param link: An instance of the Link class that represents the link that the points are related to.
        :return: A ContactPointsList instance that represents the contact points of the link.
        """
        return self.get_points_of_body(link)

    def get_points_of_body(self, body: PhysicalBody) -> ContactPointsList:
        """
        Get the points of the body.

        :param body: An instance of the PhysicalBody class that represents the body that the points are related to.
        :return: A ContactPointsList instance that represents the contact points of the body.
        """
        return ContactPointsList([point for point in self if body == point.body_b])

    def get_objects_that_got_removed(self, previous_points: ContactPointsList) -> List[Object]:
        """
        Return the object that is not in the current points list but was in the initial points list.

        :param previous_points: The initial points list.
        :return: A list of Object instances that represent the objects that got removed.
        """
        initial_objects_in_contact = previous_points.get_objects_that_have_points()
        current_objects_in_contact = self.get_objects_that_have_points()
        return [obj for obj in initial_objects_in_contact if obj not in current_objects_in_contact]

    def get_new_objects(self, previous_points: ContactPointsList) -> List[Object]:
        """
        Return the object that is not in the initial points list but is in the current points list.

        :param previous_points: The initial points list.
        :return: A list of Object instances that represent the new objects.
        """
        initial_objects_in_contact = previous_points.get_objects_that_have_points()
        current_objects_in_contact = self.get_objects_that_have_points()
        return [obj for obj in current_objects_in_contact if obj not in initial_objects_in_contact]

    def is_object_in_the_list(self, obj: Object) -> bool:
        """
        Check if the object is one of the objects that have points in the list.

        :param obj: An instance of the Object class that represents the object.
        :return: True if the object is in the list, False otherwise.
        """
        return obj in self.get_objects_that_have_points()

    def get_names_of_objects_that_have_points(self) -> List[str]:
        """
        Return the names of the objects that have points in the list.

        :return: A list of strings that represent the names of the objects that have points in the list.
        """
        return [obj.name for obj in self.get_objects_that_have_points()]

    def get_objects_that_have_points(self) -> List[Object]:
        """
        Return the objects that have points in the list.

        :return: A list of Object instances that represent the objects that have points in the list.
        """
        return list({point.body_b.parent_entity for point in self})

    def __str__(self):
        return f"ContactPointsList: {', '.join([point.__str__() for point in self])}"

    def __repr__(self):
        return self.__str__()


ClosestPointsList = ContactPointsList
"""
The list of closest points which has same structure as ContactPointsList.
"""


@dataclass
class TextAnnotation:
    """
    Dataclass for storing text annotations that can be displayed in the simulation.
    """
    text: str
    position: List[float]
    id: int
    color: Color = field(default_factory=lambda: Color(0, 0, 0, 1))
    size: float = 0.1


@dataclass
class VirtualJoint:
    """
    A virtual (not real) joint that is most likely used for simulation purposes.
    """
    name: str
    type_: JointType
    axes: Optional[Point] = None

    @property
    def type(self):
        return self.type_

    @property
    def is_virtual(self):
        return True

    def __hash__(self):
        return hash(self.name)


class Rotations(Dict[Optional[Union[Grasp, bool]], List[float]]):
    """
    A dictionary that defines standard quaternions for different grasps and orientations. This is mainly used
    to automatically calculate all grasp descriptions of a robot gripper for the robot description.

    SIDE_ROTATIONS: The quaternions for the different approach directions (front, back, left, right)
    VERTICAL_ROTATIONS: The quaternions for the different vertical alignments, in case the object requires for
    example a top grasp
    HORIZONTAL_ROTATIONS: The quaternions for the different horizontal alignments, in case the gripper needs to roll
    90°
    """
    SIDE_ROTATIONS = {
        Grasp.FRONT: [0, 0, 0, 1],
        Grasp.BACK: [0, 0, 1, 0],
        Grasp.LEFT: [0, 0, -math.sqrt(2) / 2, math.sqrt(2) / 2],
        Grasp.RIGHT: [0, 0, math.sqrt(2) / 2, math.sqrt(2) / 2],
    }

    VERTICAL_ROTATIONS = {
        None: [0, 0, 0, 1],
        Grasp.TOP: [0, math.sqrt(2) / 2, 0, math.sqrt(2) / 2],
        Grasp.BOTTOM: [0, -math.sqrt(2) / 2, 0, math.sqrt(2) / 2],
    }

    HORIZONTAL_ROTATIONS = {
        False: [0, 0, 0, 1],
        True: [math.sqrt(2) / 2, 0, 0, math.sqrt(2) / 2],
    }


@dataclass
class VirtualMobileBaseJoints:
    """
    Dataclass for storing the names, types and axes of the virtual mobile base joints of a mobile robot.
    """

    translation_x: Optional[VirtualJoint] = VirtualJoint(VirtualMobileBaseJointName.LINEAR_X.value,
                                                         JointType.PRISMATIC,
                                                         Point(x=1.0, y=0.0, z=0.0))
    translation_y: Optional[VirtualJoint] = VirtualJoint(VirtualMobileBaseJointName.LINEAR_Y.value,
                                                         JointType.PRISMATIC,
                                                         Point(x=0.0, y=1.0, z=0.0))
    angular_z: Optional[VirtualJoint] = VirtualJoint(VirtualMobileBaseJointName.ANGULAR_Z.value,
                                                     JointType.REVOLUTE,
                                                     Point(x=0.0, y=0.0, z=1.0))

    @property
    def names(self) -> List[str]:
        """
        Return the names of the virtual mobile base joints.
        """
        return [getattr(self, field.name).name for field in fields(self)]

    def get_types(self) -> Dict[str, JointType]:
        """
        Return the joint types of the virtual mobile base joints.
        """
        return {getattr(self, field.name).name: getattr(self, field.name).type_ for field in fields(self)}

    def get_axes(self) -> Dict[str, Point]:
        """
        Return the axes (i.e. The axis on which the joint moves) of the virtual mobile base joints.
        """
        return {getattr(self, field.name).name: getattr(self, field.name).axes for field in fields(self)}


@dataclass
class MultiverseMetaData:
    """Meta data for the Multiverse Client, the simulation_name should be non-empty and unique for each simulation"""
    world_name: str = "world"
    simulation_name: str = "cram"
    length_unit: str = "m"
    angle_unit: str = "rad"
    mass_unit: str = "kg"
    time_unit: str = "s"
    handedness: str = "rhs"


@dataclass
class RayResult:
    """
    A dataclass to store the ray result. The ray result contains the body name that the ray intersects with and the
    distance from the ray origin to the intersection point.
    """
    obj_id: int
    """
    The object id of the body that the ray intersects with.
    """
    link_id: int = -1
    """
    The link id of the body that the ray intersects with, -1 if root link or None.
    """
    _hit_fraction: Optional[float] = None  # TODO: Not sure of definition
    """
    The fraction of the ray length at which the intersection point is located a range in [0, 1].
    """
    hit_position: Optional[List[float]] = None
    """
    The intersection point in cartesian world coordinates.
    """
    hit_normal: Optional[List[float]] = None
    """
    The normal at the intersection point in cartesian world coordinates.
    """
    distance: Optional[float] = None
    """
    The distance from the ray origin to the intersection point.
    """

    @property
    def intersected(self) -> bool:
        """
        Check if the ray intersects with a body.
        return: Whether the ray intersects with a body.
        """
        #if not self.obj_id:
        #    logwarn("obj_id should be available to check if the ray intersects with a body,"
        #            "It appears that the ray result is not valid.")
        return self.obj_id != -1

    @property
    def hit_fraction(self) -> Optional[float]:
        if not self._hit_fraction and self.obj_id == -1:
            return 1.0
        return self._hit_fraction

    @hit_fraction.setter
    def hit_fraction(self, value: float):
        self._hit_fraction = value

    def update_distance(self, from_position: List[float], to_position: Optional[List[float]] = None) -> float:
        """
        The distance from the ray origin to the intersection point.
        """
        if self.hit_position:
            self.distance = float(np.linalg.norm(np.array(self.hit_position) - np.array(from_position)))
            if not self.hit_fraction:
                self.hit_fraction = self.distance / np.linalg.norm(np.array(to_position) - np.array(from_position))
            return self.distance
        elif not self.hit_fraction or not to_position:
            raise ValueError(f"Either hit_position or (to_position and hit_fraction)"
                             f" should be available to calculate distance,"
                             f" given hit_fraction: {self.hit_fraction}, to_position: {to_position}")
        return np.linalg.norm(np.array(to_position) - np.array(from_position)) * self.hit_fraction


@deprecated("Use RayResult instead")
@dataclass
class MultiverseRayResult:
    """
    A dataclass to store the ray result. The ray result contains the body name that the ray intersects with and the
    distance from the ray origin to the intersection point.
    """
    body_name: str
    distance: float

    def intersected(self) -> bool:
        """
        Check if the ray intersects with a body.
        return: Whether the ray intersects with a body.
        """
        return self.distance >= 0 and self.body_name != ""


@dataclass
class MultiverseContactPoint:
    """
    A dataclass to store all the contact data returned from Multiverse for a single object.
    """
    body_1: str
    body_2: str
    position: List[float]
    normal: List[float]


@dataclass
class ReasoningResult:
    """
    Result of a reasoning result of knowledge source
    """
    success: bool
    reasoned_parameter: Dict[str, Any] = field(default_factory=dict)



@dataclass
class FrozenObject:

    name: str
    """
    Name of this Object
    """
    concept: Type[PhysicalObject]
    """
    The Concept of the Object as the PyCRAP concept
    """
    path: Optional[str] = None
    """
    The path to the source file
    """
    description: Optional[ObjectDescription] = None
    """
    The description of the object, this is a combination of links and joints
    """
    pose: Optional[PoseStamped] = field(default_factory=PoseStamped)
    """
    The pose at which this object is placed
    """
    links: Optional[Dict[str, FrozenLink]] = None
    """
    A dictionary with the link name as key and the link object as value
    """
    joints: Optional[Dict[str, FrozenJoint]] = None
    """
    A dictionary of all joints, with the joint name as key and the joint object as value
    """


@dataclass(frozen=True)
class FrozenLink:
    name: str
    """
    Name of this FrozenLink
    """
    pose: PoseStamped
    """
    Pose of this Link in the world frame
    """
    geometry: Union[VisualShape, List[VisualShape]]
    """
    The geometry of this link
    """


@dataclass(frozen=True)
class FrozenJoint:
    name: str
    """
    Name of this FrozenJoints
    """
    type: JointType
    """
    The type of this joint
    """
    children: Sequence[str]
    """
    A sequence of the names of all children
    """
    parent: Optional[str]
    """
    The name of the parent joint or None if there is no parent joint
    """
    state: float
    """
    State of the joint
    """


import threading
import time
import unittest
from pycram.designators.action_designator import *
from pycram.designators.object_designator import BelieveObject
from pycram.datastructures.enums import ObjectType, DetectionTechnique, TaskStatus
from pycram.failure_handling import RetryMonitor
from pycram.fluent import Fluent
from pycram.failures import PlanFailure, NotALanguageExpression
from pycram.datastructures.pose import PoseStamped
from pycram.language import SequentialPlan, ParallelPlan, TryInOrderPlan, TryAllPLan, MonitorPlan, MonitorNode, \
    SequentialNode, RepeatPlan, CodePlan, TryAllNode
from pycram.process_module import simulated_robot
from pycram.testing import BulletWorldTestCase
from pycram.robot_description import RobotDescription
from pycrap.ontologies import Cup



class LanguageTestCase(BulletWorldTestCase):

    def test_simplify_tree(self):
        act = NavigateActionDescription(PoseStamped())
        act2 = MoveTorsoActionDescription(TorsoState.HIGH)
        act3 = DetectActionDescription(DetectionTechnique.TYPES)

        plan = SequentialPlan(act, SequentialPlan(act2, act3))
        self.assertEqual(len(plan.root.children), 3)
        self.assertEqual(plan.root.children[0].children, [])

    def test_sequential_construction(self):
        act = NavigateActionDescription(PoseStamped())
        act2 = MoveTorsoActionDescription(TorsoState.HIGH)
        act3 = DetectActionDescription(DetectionTechnique.TYPES)

        plan = SequentialPlan(act, act2, act3)
        self.assertTrue(isinstance(plan, SequentialPlan))
        self.assertEqual(len(plan.root.children), 3)

    def test_parallel_construction(self):
        act = NavigateActionDescription(PoseStamped())
        act2 = MoveTorsoActionDescription(TorsoState.HIGH)
        act3 = DetectActionDescription(DetectionTechnique.TYPES)

        plan = ParallelPlan(act, act2, act3)
        self.assertTrue(isinstance(plan, ParallelPlan))
        self.assertEqual(len(plan.root.children), 3)

    def test_try_in_order_construction(self):
        act = NavigateActionDescription(PoseStamped())
        act2 = MoveTorsoActionDescription(TorsoState.HIGH)
        act3 = DetectActionDescription(DetectionTechnique.TYPES)

        plan = TryInOrderPlan(act, act2, act3)
        self.assertTrue(isinstance(plan, TryInOrderPlan))
        self.assertEqual(len(plan.root.children), 3)

    def test_try_all_construction(self):
        act = NavigateActionDescription(PoseStamped())
        act2 = MoveTorsoActionDescription(TorsoState.HIGH)
        act3 = DetectActionDescription(DetectionTechnique.TYPES)

        plan = TryAllPLan(act, act2, act3)
        self.assertTrue(TryAllNode, type(plan.root))
        self.assertTrue(isinstance(plan, TryAllPLan))
        self.assertEqual(len(plan.root.children), 3)

    def test_combination_construction(self):
        act = NavigateActionDescription(PoseStamped())
        act2 = MoveTorsoActionDescription(TorsoState.HIGH)
        act3 = DetectActionDescription(DetectionTechnique.TYPES)

        plan = ParallelPlan( SequentialPlan(act, act2), act3)
        self.assertTrue(isinstance(plan, ParallelPlan))
        self.assertEqual(len(plan.root.children), 2)
        self.assertTrue(isinstance(plan.root.children[0], SequentialNode))

    def test_pickup_par_construction(self):
        act = NavigateActionDescription(PoseStamped())
        act2 = PickUpActionDescription(BelieveObject(names=["milk"]), ["left"], ["front"])

        self.assertRaises(AttributeError, lambda: ParallelPlan(act, act2))

    def test_pickup_try_all_construction(self):
        act = NavigateActionDescription(PoseStamped())
        act2 = PickUpActionDescription(BelieveObject(names=["milk"]), ["left"], ["front"])

        self.assertRaises(AttributeError, lambda: TryAllPLan(act, act2))

    def test_monitor_construction(self):
        act = ParkArmsActionDescription(Arms.BOTH)
        act2 = MoveTorsoActionDescription(TorsoState.HIGH)

        def monitor_func():
            time.sleep(1)
            return True

        plan = MonitorPlan(monitor_func, SequentialPlan(act, act2))
        self.assertEqual(len(plan.root.children), 1)
        self.assertTrue(isinstance(plan.root, MonitorNode))

    def test_retry_monitor_construction(self):
        act = ParkArmsActionDescription(Arms.BOTH)
        act2 = MoveTorsoActionDescription(TorsoState.HIGH)

        def monitor_func():
            time.sleep(1)
            return True

        def recovery1():
            return

        recovery = {NotALanguageExpression: recovery1}

        subplan = MonitorPlan(monitor_func, SequentialPlan(act, act2))
        plan = RetryMonitor(subplan, max_tries=6, recovery=recovery)
        self.assertEqual(len(plan.recovery), 1)
        self.assertIsInstance(plan.plan, MonitorPlan)

    def test_retry_monitor_tries(self):
        def raise_failure():
            raise PlanFailure
        tries_counter = 0

        def monitor_func():
            nonlocal tries_counter
            tries_counter += 1
            return True
        act2 = MoveTorsoActionDescription([TorsoState.HIGH])
        fail = CodePlan(raise_failure)
        counter = CodePlan(monitor_func)

        subplan = SequentialPlan(counter, fail)
        plan = RetryMonitor(subplan, max_tries=6)
        self.assertRaises(PlanFailure, plan.perform)

        self.assertEqual(tries_counter, 6)

    def test_retry_monitor_recovery(self):
        recovery1_counter = 0
        recovery2_counter = 0

        def monitor_func():
            if not hasattr(monitor_func, 'tries_counter'):
                monitor_func.tries_counter = 0
            if monitor_func.tries_counter % 2:
                monitor_func.tries_counter += 1
                raise NotALanguageExpression
            monitor_func.tries_counter += 1
            raise PlanFailure

        def recovery1():
            nonlocal recovery1_counter
            recovery1_counter += 1

        def recovery2():
            nonlocal recovery2_counter
            recovery2_counter += 1

        recovery = {NotALanguageExpression: recovery1,
                    PlanFailure: recovery2}

        code = CodePlan(monitor_func)
        subplan = SequentialPlan(code)
        plan = RetryMonitor(subplan, max_tries=6, recovery=recovery)
        try:
            plan.perform()
        except PlanFailure as e:
            pass
        self.assertEqual(2, recovery1_counter)
        self.assertEqual(3, recovery2_counter)

    def test_repeat_construction(self):
        act = ParkArmsActionDescription([Arms.BOTH])
        act2 = MoveTorsoActionDescription([TorsoState.HIGH])

        plan = RepeatPlan(5, SequentialPlan(act, act2))
        self.assertEqual(len(plan.root.children), 1)
        self.assertTrue(isinstance(plan.root.children[0], SequentialNode))

    def test_repeat_construction_error(self):
        act = ParkArmsActionDescription([Arms.BOTH])
        act2 = MoveTorsoActionDescription([TorsoState.HIGH])
        park = ParkArmsActionDescription([Arms.BOTH])

        self.assertRaises(AttributeError, lambda: RepeatPlan(park, SequentialPlan(act, act2)))

    def test_perform_desig(self):
        act = NavigateActionDescription([PoseStamped.from_list([0.3, 0.3, 0])])
        act2 = MoveTorsoActionDescription([TorsoState.HIGH])
        act3 = ParkArmsActionDescription([Arms.BOTH])

        plan = SequentialPlan(act, act2, act3)
        with simulated_robot:
            plan.perform()
        self.assertEqual(self.robot.get_pose(), PoseStamped.from_list([0.3, 0.3, 0]))
        self.assertEqual(self.robot.get_joint_position("torso_lift_joint"), 0.3)
        for joint, pose in RobotDescription.current_robot_description.get_static_joint_chain("right", StaticJointState.Park).items():
            self.assertEqual(self.world.robot.get_joint_position(joint), pose)
        for joint, pose in RobotDescription.current_robot_description.get_static_joint_chain("left", StaticJointState.Park).items():
            self.assertEqual(self.world.robot.get_joint_position(joint), pose)


    def test_perform_parallel(self):

        def check_thread_id(main_id):
            self.assertNotEqual(main_id, threading.get_ident())
        act = CodePlan(check_thread_id, {"main_id": threading.get_ident()})
        act2 = CodePlan(check_thread_id, {"main_id": threading.get_ident()})
        act3 = CodePlan(check_thread_id, {"main_id": threading.get_ident()})

        plan = ParallelPlan(act, act2, act3)
        with simulated_robot:
            plan.perform()

    def test_perform_repeat(self):
        test_var = Fluent(0)

        def inc(var):
            var.set_value(var.get_value() + 1)
        plan = RepeatPlan(10, CodePlan(lambda: inc(test_var)))
        with simulated_robot:
            plan.perform()
        self.assertEqual(10, test_var.get_value())

    def test_exception_sequential(self):
        def raise_except():
            raise PlanFailure
        act = NavigateActionDescription([PoseStamped()])
        code = CodePlan(raise_except)

        plan = SequentialPlan(act, code)
        def perform_plan():
            with simulated_robot:
                _ = plan.perform()
        self.assertRaises(PlanFailure, perform_plan)
        self.assertEqual(2, len(plan.root.children))
        self.assertEqual(TaskStatus.FAILED, plan.root.status)

    def test_exception_try_in_order(self):
        def raise_except():
            raise PlanFailure
        act = NavigateActionDescription([PoseStamped()])
        code = CodePlan(raise_except)

        plan = TryInOrderPlan(act, code)
        with simulated_robot:
            _ = plan.perform()
        self.assertEqual(2, len(plan.root.children))
        self.assertEqual(TaskStatus.SUCCEEDED, plan.root.status)

    def test_exception_parallel(self):
        def raise_except():
            raise PlanFailure
        act = NavigateActionDescription([PoseStamped()])
        code = CodePlan(raise_except)

        plan = ParallelPlan(act, code)
        with simulated_robot:
            _ = plan.perform()
        self.assertEqual(PlanFailure, type(plan.root.reason))
        self.assertEqual(plan.root.status, TaskStatus.FAILED)

    def test_exception_try_all(self):
        def raise_except():
            raise PlanFailure
        act = NavigateActionDescription([PoseStamped()])
        code = CodePlan(raise_except)

        plan = TryAllPLan(act, code)
        with simulated_robot:
            _ = plan.perform()

        self.assertEqual(TryAllNode, type(plan.root))
        self.assertEqual(TaskStatus.SUCCEEDED, plan.root.status)


if __name__ == '__main__':
    unittest.main()



import unittest
import os
from unittest.mock import patch, MagicMock
from pycram.datastructures.enums import FilterConfig
try:
    from geometry_msgs.msg import WrenchStamped
    from pycram.ros_utils.force_torque_sensor import ForceTorqueSensor, ForceTorqueSensorSimulated
except ImportError:
    pass

@unittest.skipIf("ROS_VERSION" not in os.environ, "ROS is not available")
class ForceTorqueSensorTestCase(unittest.TestCase):

    @patch('pycram.ros_utils.force_torque_sensor.World')
    @patch('pycram.ros_utils.force_torque_sensor.create_publisher')
    def test_initialization_simulated_sensor(self, mock_create_publisher, mock_world):
        mock_world.current_world.robot.joint_name_to_id = {'joint1': 1}
        sensor = ForceTorqueSensorSimulated('joint1')
        self.assertEqual('joint1', 'joint1')
        self.assertIsNotNone(sensor.fts_pub)
        sensor._stop_publishing()

    @patch('pycram.ros_utils.force_torque_sensor.World')
    @patch('pycram.ros_utils.force_torque_sensor.create_publisher')
    def test_initialization_simulated_sensor_invalid_joint(self, mock_create_publisher, mock_world):
        mock_world.current_world.robot.joint_name_to_id = {}
        with self.assertRaises(RuntimeError):
            ForceTorqueSensorSimulated('invalid_joint')

    def test_initialization_force_torque_sensor(self):
        sensor = ForceTorqueSensor('hsrb')
        self.assertEqual(sensor.robot_name, 'hsrb')
        self.assertEqual(sensor.wrench_topic_name, '/hsrb/wrist_wrench/compensated')

    def test_get_last_value(self):
        sensor = ForceTorqueSensor('hsrb')
        mock_data = MagicMock(spec=WrenchStamped)
        sensor.whole_data = {sensor.filtered: [mock_data], sensor.unfiltered: [mock_data]}
        self.assertEqual(sensor.get_last_value(), mock_data)

    def test_get_derivative(self):
        sensor = ForceTorqueSensor('hsrb')
        mock_data1 = MagicMock(spec=WrenchStamped)
        mock_data2 = MagicMock(spec=WrenchStamped)
        sensor.whole_data = {sensor.filtered: [mock_data1, mock_data2], sensor.unfiltered: [mock_data1, mock_data2]}
        derivative = sensor.get_derivative()
        self.assertIsInstance(derivative, WrenchStamped)

if __name__ == '__main__':
    unittest.main()


from __future__ import annotations
from datetime import timedelta
from time import sleep, time

import numpy as np
from typing_extensions import Any, Callable, Optional, Union, Iterable, Dict, TYPE_CHECKING, Tuple

from .error_checkers import ErrorChecker, PoseErrorChecker, PositionErrorChecker, \
    OrientationErrorChecker, SingleValueErrorChecker
from ..datastructures.enums import JointType
from ..ros import logerr, logwarn

if TYPE_CHECKING:
    from ..datastructures.world import World
    from ..world_concepts.world_object import Object
    from ..datastructures.pose import PoseStamped
    from ..description import ObjectDescription

    Joint = ObjectDescription.Joint
    Link = ObjectDescription.Link

OptionalArgCallable = Union[Callable[[], Any], Callable[[Any], Any]]


class GoalValidator:
    """
    A class to validate the goal by tracking the goal achievement progress.
    """

    raise_error: Optional[bool] = False
    """
    Whether to raise an error if the goal is not achieved.
    """
    total_wait_time: Optional[timedelta] = None
    """
    The total wait time that was spent waiting for the goal to be achieved.
    """

    def __init__(self, error_checker: ErrorChecker, current_value_getter: OptionalArgCallable,
                 acceptable_percentage_of_goal_achieved: Optional[float] = 0.8):
        """
        Initialize the goal validator.

        :param error_checker: The error checker.
        :param current_value_getter: The current value getter function which takes an optional input and returns the
        current value.
        :param acceptable_percentage_of_goal_achieved: The acceptable percentage of goal achieved, if given, will be
        used to check if this percentage is achieved instead of the complete goal.
        """
        self.error_checker: ErrorChecker = error_checker
        self.current_value_getter: Callable[[Optional[Any]], Any] = current_value_getter
        self.acceptable_percentage_of_goal_achieved: Optional[float] = acceptable_percentage_of_goal_achieved
        self.goal_value: Optional[Any] = None
        self.initial_error: Optional[np.ndarray] = None
        self.current_value_getter_input: Optional[Any] = None

    def register_goal_and_wait_until_achieved(self, goal_value: Any,
                                              current_value_getter_input: Optional[Any] = None,
                                              initial_value: Optional[Any] = None,
                                              acceptable_error: Optional[Union[float, Iterable[float]]] = None,
                                              max_wait_time: Optional[float] = 1,
                                              time_per_read: Optional[float] = 0.01) -> None:
        """
        Register the goal value and wait until the target is reached.

        :param goal_value: The goal value.
        :param current_value_getter_input: The values that are used as input to the current value getter.
        :param initial_value: The initial value.
        :param acceptable_error: The acceptable error.
        :param max_wait_time: The maximum time to wait.
        :param time_per_read: The time to wait between each read.
        """
        self.register_goal(goal_value, current_value_getter_input, initial_value, acceptable_error)
        self.wait_until_goal_is_achieved(max_wait_time, time_per_read)

    def wait_until_goal_is_achieved(self, max_wait_time: Optional[timedelta] = timedelta(seconds=2),
                                    time_per_read: Optional[timedelta] = timedelta(milliseconds=10)) -> None:
        """
        Wait until the target is reached.

        :param max_wait_time: The maximum time to wait.
        :param time_per_read: The time to wait between each read.
        """
        if self.goal_value is None:
            logwarn("Goal value is None, skipping waiting for goal to be achieved")
            return  # Skip if goal value is None
        start_time = time()
        current = self.current_value
        while not self.goal_achieved:
            self.total_wait_time = timedelta(seconds=time() - start_time)
            if self.total_wait_time > max_wait_time:
                msg = f"Failed to achieve goal from initial error {self.initial_error} with" \
                      f" goal {self.goal_value} within {max_wait_time.total_seconds()}" \
                      f" seconds, the current value is {current}, error is {self.current_error}, percentage" \
                      f" of goal achieved is {self.percentage_of_goal_achieved}"
                if self.raise_error:
                    logerr(msg)
                    raise TimeoutError(msg)
                else:
                    logwarn(msg)
                    break
            sleep(time_per_read.total_seconds())
            current = self.current_value

    def reset(self) -> None:
        """
        Reset the goal validator.
        """
        self.goal_value = None
        self.initial_error = None
        self.current_value_getter_input = None
        self.error_checker.reset()

    @property
    def _acceptable_error(self) -> np.ndarray:
        """
        The acceptable error.
        """
        if self.error_checker.is_iterable:
            return self.tiled_acceptable_error
        else:
            return self.acceptable_error

    @property
    def acceptable_error(self) -> np.ndarray:
        """
        The acceptable error.
        """
        return self.error_checker.acceptable_error

    @property
    def tiled_acceptable_error(self) -> Optional[np.ndarray]:
        """
        The tiled acceptable error.
        """
        return self.error_checker.tiled_acceptable_error

    def register_goal(self, goal_value: Any,
                      current_value_getter_input: Optional[Any] = None,
                      initial_value: Optional[Any] = None,
                      acceptable_error: Optional[Union[float, Iterable[float]]] = None):
        """
        Register the goal value.

        :param goal_value: The goal value.
        :param current_value_getter_input: The values that are used as input to the current value getter.
        :param initial_value: The initial value.
        :param acceptable_error: The acceptable error.
        """
        if goal_value is None or (hasattr(goal_value, '__len__') and len(goal_value) == 0):
            return  # Skip if goal value is None or empty
        self.goal_value = goal_value
        self.current_value_getter_input = current_value_getter_input
        self.update_initial_error(goal_value, initial_value=initial_value)
        self.error_checker.update_acceptable_error(acceptable_error, self.initial_error)

    def update_initial_error(self, goal_value: Any, initial_value: Optional[Any] = None) -> None:
        """
        Calculate the initial error.

        :param goal_value: The goal value.
        :param initial_value: The initial value.
        """
        if initial_value is None:
            self.initial_error: np.ndarray = self.current_error
        else:
            self.initial_error: np.ndarray = self.calculate_error(goal_value, initial_value)

    @property
    def current_value(self) -> Any:
        """
        The current value of the monitored variable.
        """
        if self.current_value_getter_input is not None:
            return self.current_value_getter(self.current_value_getter_input)
        else:
            try:
                return self.current_value_getter()
            except Exception as e:
                logerr(f"Error while getting current value: {e}")
                return None

    @property
    def current_error(self) -> np.ndarray:
        """
        The current error.
        """
        return self.calculate_error(self.goal_value, self.current_value)

    def calculate_error(self, value_1: Any, value_2: Any) -> np.ndarray:
        """
        Calculate the error between two values.

        :param value_1: The first value.
        :param value_2: The second value.
        :return: The error.
        """
        return np.array(self.error_checker.calculate_error(value_1, value_2)).flatten()

    @property
    def percentage_of_goal_achieved(self) -> float:
        """
        The relative (relative to the acceptable error) achieved percentage of goal.
        """
        percent_array = 1 - self.relative_current_error / self.relative_initial_error
        percent_array_filtered = percent_array[self.relative_initial_error > self._acceptable_error]
        if len(percent_array_filtered) == 0:
            return 1
        else:
            return np.mean(percent_array_filtered)

    @property
    def actual_percentage_of_goal_achieved(self) -> float:
        """
        The percentage of goal achieved.
        """
        percent_array = 1 - self.current_error / np.maximum(self.initial_error, 1e-3)
        percent_array_filtered = percent_array[self.initial_error > self._acceptable_error]
        if len(percent_array_filtered) == 0:
            return 1
        else:
            return np.mean(percent_array_filtered)

    @property
    def relative_current_error(self) -> np.ndarray:
        """
        The relative current error (relative to the acceptable error).
        """
        return self.get_relative_error(self.current_error, threshold=0)

    @property
    def relative_initial_error(self) -> np.ndarray:
        """
        The relative initial error (relative to the acceptable error).
        """
        if self.initial_error is None:
            self.update_initial_error(self.goal_value)
        return np.maximum(self.initial_error, 1e-3)

    def get_relative_error(self, error: Any, threshold: Optional[float] = 1e-3) -> np.ndarray:
        """
        Get the relative error by comparing the error with the acceptable error and filtering out the errors that are
        less than the threshold.

        :param error: The error.
        :param threshold: The threshold.
        :return: The relative error.
        """
        return np.maximum(error - self._acceptable_error, threshold)

    @property
    def goal_achieved(self) -> bool:
        """
        Check if the goal is achieved.
        """
        if self.acceptable_percentage_of_goal_achieved is None:
            return self.is_current_error_acceptable
        else:
            return self.percentage_of_goal_achieved >= self.acceptable_percentage_of_goal_achieved

    @property
    def is_current_error_acceptable(self) -> bool:
        """
        Check if the error is acceptable.
        """
        return self.error_checker.is_error_acceptable(self.current_value, self.goal_value)

    @property
    def goal_not_achieved_message(self):
        """
        Message to be displayed when the goal is not achieved.
        """
        return f"Goal not achieved, current value: {self.current_value}, error: {self.current_error}, " \
               f"percentage of goal achieved: {self.percentage_of_goal_achieved}, " \
               f"with initial error: {self.initial_error}, goal: {self.goal_value}"


class PoseGoalValidator(GoalValidator):
    """
    A class to validate the pose goal by tracking the goal achievement progress.
    """

    def __init__(self, current_pose_getter: OptionalArgCallable = None,
                 acceptable_error: Union[Tuple[float], Iterable[Tuple[float]]] = (1e-3, np.pi / 180),
                 acceptable_percentage_of_goal_achieved: Optional[float] = 0.8,
                 is_iterable: Optional[bool] = False):
        """
        Initialize the pose goal validator.

        :param current_pose_getter: The current pose getter function which takes an optional input and returns the
        current pose.
        :param acceptable_error: The acceptable error.
        :param acceptable_percentage_of_goal_achieved: The acceptable percentage of goal achieved.
        """
        super().__init__(PoseErrorChecker(acceptable_error, is_iterable=is_iterable), current_pose_getter,
                         acceptable_percentage_of_goal_achieved=acceptable_percentage_of_goal_achieved)


class MultiPoseGoalValidator(PoseGoalValidator):
    """
    A class to validate the multi-pose goal by tracking the goal achievement progress.
    """

    def __init__(self, current_poses_getter: OptionalArgCallable = None,
                 acceptable_error: Union[Tuple[float], Iterable[Tuple[float]]] = (1e-2, 5 * np.pi / 180),
                 acceptable_percentage_of_goal_achieved: Optional[float] = 0.8):
        """
        Initialize the multi-pose goal validator.

        :param current_poses_getter: The current poses getter function which takes an optional input and returns the
        current poses.
        :param acceptable_error: The acceptable error.
        :param acceptable_percentage_of_goal_achieved: The acceptable percentage of goal achieved.
        """
        super().__init__(current_poses_getter, acceptable_error, acceptable_percentage_of_goal_achieved,
                         is_iterable=True)


class PositionGoalValidator(GoalValidator):
    """
    A class to validate the position goal by tracking the goal achievement progress.
    """

    def __init__(self, current_position_getter: OptionalArgCallable = None,
                 acceptable_error: Optional[float] = 1e-3,
                 acceptable_percentage_of_goal_achieved: Optional[float] = 0.8,
                 is_iterable: Optional[bool] = False):
        """
        Initialize the position goal validator.

        :param current_position_getter: The current position getter function which takes an optional input and
         returns the current position.
        :param acceptable_error: The acceptable error.
        :param acceptable_percentage_of_goal_achieved: The acceptable percentage of goal achieved.
        :param is_iterable: Whether it is a sequence of position vectors.
        """
        super().__init__(PositionErrorChecker(acceptable_error, is_iterable=is_iterable), current_position_getter,
                         acceptable_percentage_of_goal_achieved=acceptable_percentage_of_goal_achieved)


class MultiPositionGoalValidator(PositionGoalValidator):
    """
    A class to validate the multi-position goal by tracking the goal achievement progress.
    """

    def __init__(self, current_positions_getter: OptionalArgCallable = None,
                 acceptable_error: Optional[float] = 1e-3,
                 acceptable_percentage_of_goal_achieved: Optional[float] = 0.8):
        """
        Initialize the multi-position goal validator.

        :param current_positions_getter: The current positions getter function which takes an optional input and
         returns the current positions.
        :param acceptable_error: The acceptable error.
        :param acceptable_percentage_of_goal_achieved: The acceptable percentage of goal achieved.
        """
        super().__init__(current_positions_getter, acceptable_error, acceptable_percentage_of_goal_achieved,
                         is_iterable=True)


class OrientationGoalValidator(GoalValidator):
    """
    A class to validate the orientation goal by tracking the goal achievement progress.
    """

    def __init__(self, current_orientation_getter: OptionalArgCallable = None,
                 acceptable_error: Optional[float] = np.pi / 180,
                 acceptable_percentage_of_goal_achieved: Optional[float] = 0.8,
                 is_iterable: Optional[bool] = False):
        """
        Initialize the orientation goal validator.

        :param current_orientation_getter: The current orientation getter function which takes an optional input and
         returns the current orientation.
        :param acceptable_error: The acceptable error.
        :param acceptable_percentage_of_goal_achieved: The acceptable percentage of goal achieved.
        :param is_iterable: Whether it is a sequence of quaternions.
        """
        super().__init__(OrientationErrorChecker(acceptable_error, is_iterable=is_iterable), current_orientation_getter,
                         acceptable_percentage_of_goal_achieved=acceptable_percentage_of_goal_achieved)


class MultiOrientationGoalValidator(OrientationGoalValidator):
    """
    A class to validate the multi-orientation goal by tracking the goal achievement progress.
    """

    def __init__(self, current_orientations_getter: OptionalArgCallable = None,
                 acceptable_error: Optional[float] = np.pi / 180,
                 acceptable_percentage_of_goal_achieved: Optional[float] = 0.8):
        """
        Initialize the multi-orientation goal validator.

        :param current_orientations_getter: The current orientations getter function which takes an optional input and
         returns the current orientations.
        :param acceptable_error: The acceptable error.
        :param acceptable_percentage_of_goal_achieved: The acceptable percentage of goal achieved.
        """
        super().__init__(current_orientations_getter, acceptable_error, acceptable_percentage_of_goal_achieved,
                         is_iterable=True)


class JointPositionGoalValidator(GoalValidator):
    """
    A class to validate the joint position goal by tracking the goal achievement progress.
    """

    def __init__(self, current_position_getter: OptionalArgCallable = None,
                 acceptable_error: Optional[float] = None,
                 acceptable_revolute_joint_position_error: float = np.pi / 180,
                 acceptable_prismatic_joint_position_error: float = 1e-3,
                 acceptable_percentage_of_goal_achieved: float = 0.8,
                 is_iterable: bool = False):
        """
        Initialize the joint position goal validator.

        :param current_position_getter: The current position getter function which takes an optional input and returns
         the current position.
        :param acceptable_error: The acceptable error.
        :param acceptable_revolute_joint_position_error: The acceptable orientation error.
        :param acceptable_prismatic_joint_position_error: The acceptable position error.
        :param acceptable_percentage_of_goal_achieved: The acceptable percentage of goal achieved.
        :param is_iterable: Whether it is a sequence of joint positions.
        """
        super().__init__(SingleValueErrorChecker(acceptable_error, is_iterable=is_iterable), current_position_getter,
                         acceptable_percentage_of_goal_achieved=acceptable_percentage_of_goal_achieved)
        self.acceptable_orientation_error = acceptable_revolute_joint_position_error
        self.acceptable_position_error = acceptable_prismatic_joint_position_error

    def register_goal(self, goal_value: Any, joint_type: JointType,
                      current_value_getter_input: Optional[Any] = None,
                      initial_value: Optional[Any] = None,
                      acceptable_error: Optional[float] = None):
        """
        Register the goal value.

        :param goal_value: The goal value.
        :param joint_type: The joint type (e.g. REVOLUTE, PRISMATIC).
        :param current_value_getter_input: The values that are used as input to the current value getter.
        :param initial_value: The initial value.
        :param acceptable_error: The acceptable error.
        """
        if acceptable_error is None:
            self.error_checker.acceptable_error = self.acceptable_orientation_error if joint_type == JointType.REVOLUTE \
                else self.acceptable_position_error
        super().register_goal(goal_value, current_value_getter_input, initial_value, acceptable_error)


class MultiJointPositionGoalValidator(GoalValidator):
    """
    A class to validate the multi-joint position goal by tracking the goal achievement progress.
    """

    def __init__(self, current_positions_getter: OptionalArgCallable = None,
                 acceptable_error: Optional[Iterable[float]] = None,
                 acceptable_revolute_joint_position_error: float = np.pi / 180,
                 acceptable_prismatic_joint_position_error: float = 1e-3,
                 acceptable_percentage_of_goal_achieved: float = 0.8):
        """
        Initialize the multi-joint position goal validator.

        :param current_positions_getter: The current positions getter function which takes an optional input and
         returns the current positions.
        :param acceptable_error: The acceptable error.
        :param acceptable_revolute_joint_position_error: The acceptable orientation error.
        :param acceptable_prismatic_joint_position_error: The acceptable position error.
        :param acceptable_percentage_of_goal_achieved: The acceptable percentage of goal achieved.
        """
        super().__init__(SingleValueErrorChecker(acceptable_error, is_iterable=True), current_positions_getter,
                         acceptable_percentage_of_goal_achieved)
        self.acceptable_orientation_error = acceptable_revolute_joint_position_error
        self.acceptable_position_error = acceptable_prismatic_joint_position_error

    def register_goal(self, goal_value: Any, joint_type: Iterable[JointType],
                      current_value_getter_input: Optional[Any] = None,
                      initial_value: Optional[Any] = None,
                      acceptable_error: Optional[Iterable[float]] = None):
        if acceptable_error is None:
            self.error_checker.acceptable_error = [self.acceptable_orientation_error if jt == JointType.REVOLUTE
                                                   else self.acceptable_position_error for jt in joint_type]
        super().register_goal(goal_value, current_value_getter_input, initial_value, acceptable_error)


def validate_object_pose(pose_setter_func):
    """
    A decorator to validate the object pose.

    :param pose_setter_func: The function to set the pose of the object.
    """

    def wrapper(world: World, obj: Object, pose: PoseStamped):

        if not world.current_world.conf.validate_goals:
            return pose_setter_func(world, obj, pose)

        if obj is None:
            logerr("Object should not be None")
            return False
        pose_goal_validator = PoseGoalValidator(world.get_object_pose, world.conf.get_pose_tolerance(),
                                                world.conf.acceptable_percentage_of_goal)
        pose_goal_validator.register_goal(pose, obj)

        if not pose_setter_func(world, obj, pose):
            return False

        pose_goal_validator.wait_until_goal_is_achieved()
        return True

    return wrapper


def validate_multiple_object_poses(pose_setter_func):
    """
    A decorator to validate multiple object poses.

    :param pose_setter_func: The function to set multiple poses of the objects.
    """

    def wrapper(world: World, object_poses: Dict[Object, PoseStamped]):

        if not world.current_world.conf.validate_goals:
            return pose_setter_func(world, object_poses)

        multi_pose_goal_validator = MultiPoseGoalValidator(
            lambda x: list(world.get_multiple_object_poses(x).values()),
            world.conf.get_pose_tolerance(), world.conf.acceptable_percentage_of_goal)
        multi_pose_goal_validator.register_goal(list(object_poses.values()),
                                                list(object_poses.keys()))

        if not pose_setter_func(world, object_poses):
            return False

        multi_pose_goal_validator.wait_until_goal_is_achieved()
        return True

    return wrapper


def validate_joint_position(position_setter_func):
    """
    A decorator to validate the joint position.

    :param position_setter_func: The function to set the joint position.
    """

    def wrapper(world: World, joint: Joint, position: float):

        if not world.current_world.conf.validate_goals:
            return position_setter_func(world, joint, position)

        joint_position_goal_validator = JointPositionGoalValidator(
            world.get_joint_position,
            acceptable_revolute_joint_position_error=world.conf.revolute_joint_position_tolerance,
            acceptable_prismatic_joint_position_error=world.conf.prismatic_joint_position_tolerance,
            acceptable_percentage_of_goal_achieved=world.conf.acceptable_percentage_of_goal)

        joint_type = joint.type
        joint_position_goal_validator.register_goal(position, joint_type, joint)

        if not position_setter_func(world, joint, position):
            return False

        joint_position_goal_validator.wait_until_goal_is_achieved()
        return True

    return wrapper


def validate_multiple_joint_positions(position_setter_func):
    """
    A decorator to validate the joint positions, this function does not validate the virtual joints,
    as in multiverse the virtual joints take command velocities and not positions, so after their goals
    are set, they are zeroed thus can't be validated. (They are actually validated by the robot pose in case
    of virtual mobile base joints)

    :param position_setter_func: The function to set the joint positions.
    """

    def wrapper(world: World, joint_positions: Dict[Joint, float]):
        if not world.current_world.conf.validate_goals:
            return position_setter_func(world, joint_positions)
        joint_positions_to_validate = {joint: position for joint, position in joint_positions.items()
                                       if not joint.is_virtual}
        if not joint_positions_to_validate:
            return position_setter_func(world, joint_positions)

        robot = list(joint_positions.keys())[0].object
        multi_joint_position_goal_validator = create_multiple_joint_goal_validator(robot, joint_positions_to_validate)

        if not position_setter_func(world, joint_positions):
            return False

        multi_joint_position_goal_validator.wait_until_goal_is_achieved()
        return True

    return wrapper


def create_multiple_joint_goal_validator(robot: Object,
                                         joint_positions: Union[Dict[Joint, float], Dict[str, float]]) \
        -> MultiJointPositionGoalValidator:
    """
    Validate the multiple joint goals, and wait until the goal is achieved.

    :param robot: The robot object.
    :param joint_positions: The joint positions to validate.
    """
    world = robot.world
    multi_joint_position_goal_validator = MultiJointPositionGoalValidator(
        lambda x: list(world.get_multiple_joint_positions(x).values()),
        acceptable_revolute_joint_position_error=world.conf.revolute_joint_position_tolerance,
        acceptable_prismatic_joint_position_error=world.conf.prismatic_joint_position_tolerance,
        acceptable_percentage_of_goal_achieved=world.conf.acceptable_percentage_of_goal)
    joint_objects = list(joint_positions.keys())
    if isinstance(joint_objects[0], str):
        joint_objects = [robot.joints[joint_name] for joint_name in joint_objects]
    joint_types = [joint.type for joint in joint_objects]
    multi_joint_position_goal_validator.register_goal(list(joint_positions.values()), joint_types,
                                                      joint_objects)
    return multi_joint_position_goal_validator



from abc import ABC, abstractmethod
from collections.abc import Iterable

import numpy as np
from ..tf_transformations import quaternion_multiply, quaternion_inverse
from typing_extensions import List, Union, Optional, Any, Sized, Iterable as T_Iterable, TYPE_CHECKING, Tuple

from ..datastructures.enums import JointType
if TYPE_CHECKING:
    from ..datastructures.pose import PoseStamped


class ErrorChecker(ABC):
    """
    An abstract class that resembles an error checker. It has two main methods, one for calculating the error between
    two values and another for checking if the error is acceptable.
    """
    def __init__(self, acceptable_error: Union[float, T_Iterable[float]], is_iterable: Optional[bool] = False):
        """
        Initialize the error checker.

        :param acceptable_error: The acceptable error.
        :param is_iterable: Whether the error is iterable (i.e. list of errors).
        """
        self._acceptable_error: np.ndarray = np.array(acceptable_error)
        self.tiled_acceptable_error: Optional[np.ndarray] = None
        self.is_iterable = is_iterable

    def reset(self) -> None:
        """
        Reset the error checker.
        """
        self.tiled_acceptable_error = None

    @property
    def acceptable_error(self) -> np.ndarray:
        return self._acceptable_error

    @acceptable_error.setter
    def acceptable_error(self, new_acceptable_error: Union[float, T_Iterable[float]]) -> None:
        self._acceptable_error = np.array(new_acceptable_error)

    def update_acceptable_error(self, new_acceptable_error: Optional[T_Iterable[float]] = None,
                                tile_to_match: Optional[Sized] = None,) -> None:
        """
        Update the acceptable error with a new value, and tile it to match the length of the error if needed.

        :param new_acceptable_error: The new acceptable error.
        :param tile_to_match: The iterable to match the length of the error with.
        """
        if new_acceptable_error is not None:
            self.acceptable_error = new_acceptable_error
        if tile_to_match is not None and self.is_iterable:
            self.update_tiled_acceptable_error(tile_to_match)

    def update_tiled_acceptable_error(self, tile_to_match: Sized) -> None:
        """
        Tile the acceptable error to match the length of the error.

        :param tile_to_match: The object to match the length of the error.
        :return: The tiled acceptable error.
        """
        self.tiled_acceptable_error = np.tile(self.acceptable_error.flatten(),
                                              len(tile_to_match) // self.acceptable_error.size)

    @abstractmethod
    def _calculate_error(self, value_1: Any, value_2: Any) -> Union[float, List[float]]:
        """
        Calculate the error between two values.

        :param value_1: The first value.
        :param value_2: The second value.
        :return: The error between the two values.
        """
        pass

    def calculate_error(self, value_1: Any, value_2: Any) -> Union[float, List[float]]:
        """
        Calculate the error between two values.

        :param value_1: The first value.
        :param value_2: The second value.
        :return: The error between the two values.
        """
        if self.is_iterable:
            return [self._calculate_error(v1, v2) for v1, v2 in zip(value_1, value_2)]
        else:
            return self._calculate_error(value_1, value_2)

    def is_error_acceptable(self, value_1: Any, value_2: Any) -> bool:
        """
        Check if the error is acceptable.

        :param value_1: The first value.
        :param value_2: The second value.
        :return: Whether the error is acceptable.
        """
        error = self.calculate_error(value_1, value_2)
        if self.is_iterable:
            error = np.array(error).flatten()
            if self.tiled_acceptable_error is None or\
                    len(error) != len(self.tiled_acceptable_error):
                self.update_tiled_acceptable_error(error)
            return np.all(error <= self.tiled_acceptable_error)
        else:
            return is_error_acceptable(error, self.acceptable_error)


class PoseErrorChecker(ErrorChecker):

    def __init__(self, acceptable_error: Union[Tuple[float], T_Iterable[Tuple[float]]] = (1e-3, np.pi / 180),
                 is_iterable: Optional[bool] = False):
        """
        Initialize the pose error checker.

        :param acceptable_error: The acceptable pose error (position error, orientation error).
        :param is_iterable: Whether the error is iterable (i.e. list of errors).
        """
        super().__init__(acceptable_error, is_iterable)

    def _calculate_error(self, value_1: Any, value_2: Any) -> List[float]:
        """
        Calculate the error between two poses.

        :param value_1: The first pose.
        :param value_2: The second pose.
        """
        return calculate_pose_error(value_1, value_2)


class PositionErrorChecker(ErrorChecker):

    def __init__(self, acceptable_error: Optional[float] = 1e-3, is_iterable: Optional[bool] = False):
        """
        Initialize the position error checker.

        :param acceptable_error: The acceptable position error.
        :param is_iterable: Whether the error is iterable (i.e. list of errors).
        """
        super().__init__(acceptable_error, is_iterable)

    def _calculate_error(self, value_1: Any, value_2: Any) -> float:
        """
        Calculate the error between two positions.

        :param value_1: The first position.
        :param value_2: The second position.
        :return: The error between the two positions.
        """
        return calculate_position_error(value_1, value_2)


class OrientationErrorChecker(ErrorChecker):

    def __init__(self, acceptable_error: Optional[float] = np.pi / 180, is_iterable: Optional[bool] = False):
        """
        Initialize the orientation error checker.

        :param acceptable_error: The acceptable orientation error.
        :param is_iterable: Whether the error is iterable (i.e. list of errors).
        """
        super().__init__(acceptable_error, is_iterable)

    def _calculate_error(self, value_1: Any, value_2: Any) -> float:
        """
        Calculate the error between two quaternions.

        :param value_1: The first quaternion.
        :param value_2: The second quaternion.
        :return: The error between the two quaternions.
        """
        return calculate_orientation_error(value_1, value_2)


class SingleValueErrorChecker(ErrorChecker):

    def __init__(self, acceptable_error: Optional[float] = 1e-3, is_iterable: Optional[bool] = False):
        """
        Initialize the single value error checker.

        :param acceptable_error: The acceptable error between two values.
        :param is_iterable: Whether the error is iterable (i.e. list of errors).
        """
        super().__init__(acceptable_error, is_iterable)

    def _calculate_error(self, value_1: Any, value_2: Any) -> float:
        """
        Calculate the error between two values.

        :param value_1: The first value.
        :param value_2: The second value.
        :return: The error between the two values.
        """
        return abs(value_1 - value_2)


class RevoluteJointPositionErrorChecker(SingleValueErrorChecker):

    def __init__(self, acceptable_error: Optional[float] = np.pi / 180, is_iterable: Optional[bool] = False):
        """
        Initialize the revolute joint position error checker.

        :param acceptable_error: The acceptable revolute joint position error.
        :param is_iterable: Whether the error is iterable (i.e. list of errors).
        """
        super().__init__(acceptable_error, is_iterable)


class PrismaticJointPositionErrorChecker(SingleValueErrorChecker):

    def __init__(self, acceptable_error: Optional[float] = 1e-3, is_iterable: Optional[bool] = False):
        """
        Initialize the prismatic joint position error checker.

        :param acceptable_error: The acceptable prismatic joint position error.
        :param is_iterable: Whether the error is iterable (i.e. list of errors).
        """
        super().__init__(acceptable_error, is_iterable)


class IterableErrorChecker(ErrorChecker):

    def __init__(self, acceptable_error: Optional[T_Iterable[float]] = None):
        """
        Initialize the iterable error checker.

        :param acceptable_error: The acceptable error between two values.
        """
        super().__init__(acceptable_error, True)

    def _calculate_error(self, value_1: Any, value_2: Any) -> float:
        """
        Calculate the error between two values.

        :param value_1: The first value.
        :param value_2: The second value.
        :return: The error between the two values.
        """
        return abs(value_1 - value_2)


class MultiJointPositionErrorChecker(IterableErrorChecker):

    def __init__(self, joint_types: List[JointType], acceptable_error: Optional[T_Iterable[float]] = None):
        """
        Initialize the multi-joint position error checker.

        :param joint_types: The types of the joints.
        :param acceptable_error: The acceptable error between two joint positions.
        """
        self.joint_types = joint_types
        if acceptable_error is None:
            acceptable_error = [np.pi/180 if jt == JointType.REVOLUTE else 1e-3 for jt in joint_types]
        super().__init__(acceptable_error)

    def _calculate_error(self, value_1: Any, value_2: Any) -> float:
        """
        Calculate the error between two joint positions.

        :param value_1: The first joint position.
        :param value_2: The second joint position.
        :return: The error between the two joint positions.
        """
        return calculate_joint_position_error(value_1, value_2)


def calculate_pose_error(pose_1: 'PoseStamped', pose_2: 'PoseStamped') -> List[float]:
    """
    Calculate the error between two poses.

    :param pose_1: The first pose.
    :param pose_2: The second pose.
    :return: The error between the two poses.
    """
    return [calculate_position_error(pose_1.position.to_list(), pose_2.position.to_list()),
            calculate_orientation_error(pose_1.orientation.to_list(), pose_2.orientation.to_list())]


def calculate_position_error(position_1: List[float], position_2: List[float]) -> float:
    """
    Calculate the error between two positions.

    :param position_1: The first position.
    :param position_2: The second position.
    :return: The error between the two positions.
    """
    return np.linalg.norm(np.array(position_1) - np.array(position_2))


def calculate_orientation_error(quat_1: List[float], quat_2: List[float]) -> float:
    """
    Calculate the error between two quaternions.

    :param quat_1: The first quaternion.
    :param quat_2: The second quaternion.
    :return: The error between the two quaternions.
    """
    return calculate_angle_between_quaternions(quat_1, quat_2)


def calculate_joint_position_error(joint_position_1: float, joint_position_2: float) -> float:
    """
    Calculate the error between two joint positions.

    :param joint_position_1: The first joint position.
    :param joint_position_2: The second joint position.
    :return: The error between the two joint positions.
    """
    return abs(joint_position_1 - joint_position_2)


def is_error_acceptable(error: Union[float, T_Iterable[float]],
                        acceptable_error: Union[float, T_Iterable[float]]) -> bool:
    """
    Check if the error is acceptable.

    :param error: The error.
    :param acceptable_error: The acceptable error.
    :return: Whether the error is acceptable.
    """
    if isinstance(error, Iterable):
        return all([error_i <= acceptable_error_i for error_i, acceptable_error_i in zip(error, acceptable_error)])
    else:
        return error <= acceptable_error


def calculate_angle_between_quaternions(quat_1: List[float], quat_2: List[float]) -> float:
    """
    Calculates the angle between two quaternions.

    :param quat_1: The first quaternion.
    :param quat_2: The second quaternion.
    :return: A float value that represents the angle between the two quaternions.
    """
    quat_diff = calculate_quaternion_difference(quat_1, quat_2)
    quat_diff_angle = 2 * np.arctan2(np.linalg.norm(quat_diff[0:3]), quat_diff[3])
    if quat_diff_angle > np.pi:
        quat_diff_angle = 2 * np.pi - quat_diff_angle
    return quat_diff_angle


def calculate_quaternion_difference(quat_1: List[float], quat_2: List[float]) -> List[float]:
    """
    Calculates the quaternion difference.

    :param quat_1: The quaternion of the object at the first time step.
    :param quat_2: The quaternion of the object at the second time step.
    :return: A list of float values that represent the quaternion difference.
    """
    quat_diff = quaternion_multiply(quaternion_inverse(quat_1), quat_2)
    return quat_diff
